#!/usr/bin/python
# -*- coding: utf-8 -*-
import autonetkit.log as log
from autonetkit.compilers.device.device_base import DeviceCompiler
from autonetkit.nidb import ConfigStanza


class ServerCompiler(DeviceCompiler):

    def compile(self, node):
        node.do_render = True  # turn on rendering
        phy_node = self.anm['phy'].node(node)
        ipv4_node = self.anm['ipv4'].node(node)
        super(ServerCompiler, self).compile(node)



import logging
import autonetkit.log2
ank_logger_2 = logging.getLogger("ANK2")

class AnkElement(object):

    #TODO: put this into parent __init__?
    def init_logging(self, my_type):
        try:
            self_id = str(self)
        except Exception, e:
            #TODO: log warning here
            import autonetkit.log as log
            log.warning("Unable to set per-element logger %s", e)
            self_id = ""


        log_extra={"type": my_type, "id": self_id}
        object.__setattr__(self, 'log_extra', log_extra)

        #self.log_info = partial(ank_logger_2.info, extra=extra)
        #self.log_warning = partial(ank_logger_2.warning, extra=extra)
        #self.log_error = partial(ank_logger_2.error, extra=extra)
        #self.log_critical = partial(ank_logger_2.critical, extra=extra)
        #self.log_exception = partial(ank_logger_2.exception, extra=extra)
        #self.log_debug = partial(ank_logger_2.debug, extra=extra)

    def log_info(self, message):
        ank_logger_2.info(message, extra = self.log_extra)

    def log_warning(self, message):
        ank_logger_2.warning(message, extra = self.log_extra)

    def log_error(self, message):
        ank_logger_2.error(message, extra = self.log_extra)

    def log_critical(self, message):
        ank_logger_2.critical(message, extra = self.log_extra)

    def log_exception(self, message):
        ank_logger_2.exception(message, extra = self.log_extra)

    def log_debug(self, message):
        ank_logger_2.debug(message, extra = self.log_extra)





import autonetkit.ank as ank_utils
import autonetkit.config
import autonetkit.log as log
from netaddr import IPAddress
from autonetkit.ank import sn_preflen_to_network

SETTINGS = autonetkit.config.settings


# TODO: add unit tests for each function here and in ipv6

def extract_ipv4_blocks(anm):

    # TODO: set all these blocks globally in config file, rather than repeated
    # in load, build_network, compile, etc

    from netaddr import IPNetwork
    g_in = anm['input']
    ipv4_defaults = SETTINGS["IP Addressing"]["v4"]

    # TODO: wrap these in a common function

    try:
        infra_subnet = g_in.data.ipv4_infra_subnet
        infra_prefix = g_in.data.ipv4_infra_prefix
        infra_block = sn_preflen_to_network(infra_subnet, infra_prefix)
    except Exception, e:
        infra_block = IPNetwork(
            '%s/%s' % (ipv4_defaults["infra_subnet"], ipv4_defaults["infra_prefix"]))
        if infra_subnet is None or infra_prefix is None:
            log.debug('Using default IPv4 infra_subnet %s' % infra_block)
        else:
            log.warning('Unable to obtain IPv4 infra_subnet from input graph: %s, using default %s' % (
                e, infra_block))

    try:
        loopback_subnet = g_in.data.ipv4_loopback_subnet
        loopback_prefix = g_in.data.ipv4_loopback_prefix
        loopback_block = sn_preflen_to_network(loopback_subnet,
                                               loopback_prefix)
    except Exception, e:
        loopback_block = IPNetwork(
            '%s/%s' % (ipv4_defaults["loopback_subnet"], ipv4_defaults["loopback_prefix"]))
        if loopback_subnet is None or loopback_prefix is None:
            log.debug('Using default IPv4 loopback_subnet %s' % loopback_block)
        else:
            log.warning('Unable to obtain IPv4 loopback_subnet from input graph: %s, using default %s' % (
                e, loopback_block))

    try:
        vrf_loopback_subnet = g_in.data.ipv4_vrf_loopback_subnet
        vrf_loopback_prefix = g_in.data.ipv4_vrf_loopback_prefix
        vrf_loopback_block = sn_preflen_to_network(vrf_loopback_subnet,
                                                   vrf_loopback_prefix)
    except Exception, e:
        vrf_loopback_block = IPNetwork(
            '%s/%s' % (ipv4_defaults["vrf_loopback_subnet"], ipv4_defaults["vrf_loopback_prefix"]))
        if vrf_loopback_subnet is None or vrf_loopback_prefix is None:
            log.debug('Using default IPv4 vrf_loopback_subnet %s' %
                      vrf_loopback_block)
        else:
            log.warning('Unable to obtain IPv4 vrf_loopback_subnet from input graph: %s, using default %s' % (
                e, vrf_loopback_block))

    return (infra_block, loopback_block, vrf_loopback_block)


def manual_ipv4_infrastructure_allocation(anm):
    """Applies manual IPv4 allocation"""

    import netaddr
    g_ipv4 = anm['ipv4']
    g_in = anm['input']
    log.info('Using specified IPv4 infrastructure allocation')

    for node in g_ipv4.l3devices():
        for interface in node.physical_interfaces():
            if not interface['input'].is_bound:
                continue  # unbound interface
            if not interface['ipv4'].is_bound:
                continue
            if interface['ip'].allocate is False:
                # TODO: copy interface allocate attribute across
                continue

            ip_address = netaddr.IPAddress(interface['input'
                                                     ].ipv4_address)
            prefixlen = interface['input'].ipv4_prefixlen
            interface.ip_address = ip_address
            interface.prefixlen = prefixlen
            cidr_string = '%s/%s' % (ip_address, prefixlen)
            interface.subnet = netaddr.IPNetwork(cidr_string)

    broadcast_domains = [d for d in g_ipv4 if d.broadcast_domain]

    # TODO: allow this to work with specified ip_address/subnet as well as
    # ip_address/prefixlen

    global_infra_block = None
    try:
        # Note this is only pickling up if explictly set in g_in
        infra_subnet = g_in.data.ipv4_infra_subnet
        infra_prefix = g_in.data.ipv4_infra_prefix
        global_infra_block = sn_preflen_to_network(infra_subnet, infra_prefix)
    except Exception, e:
        log.info("Unable to parse specified ipv4 infra subnets %s/%s")

    mismatched_interfaces = []
    from netaddr import IPNetwork
    for coll_dom in broadcast_domains:
        if coll_dom.allocate is False:
            continue

        # TODO: use neighbor_interfaces()
        connected_interfaces = [edge.dst_int for edge in
                                coll_dom.edges()]

        connected_interfaces = [i for i in connected_interfaces
                                if i.node.is_l3device()]

        cd_subnets = [IPNetwork('%s/%s' % (i.subnet.network,
                                           i.prefixlen)) for i in connected_interfaces
                      if i['ip'].allocate is not False]

        # mismatched_interfaces += [i for i in connected_interfaces
        # if i.
        if global_infra_block is not None:
            mismatched_interfaces += [i for i in connected_interfaces
                                      if i.ip_address not in global_infra_block]

        if len(cd_subnets) == 0:
            log.warning(
                "Collision domain %s is not connected to any nodes" % coll_dom)
            continue

        try:
            assert len(set(cd_subnets)) == 1
        except AssertionError:
            mismatch_subnets = '; '.join('%s: %s/%s' % (i,
                                                        i.subnet.network, i.prefixlen) for i in
                                         connected_interfaces)
            log.warning('Non matching subnets from collision domain %s: %s'
                        % (coll_dom, mismatch_subnets))
        else:
            coll_dom.subnet = cd_subnets[0]  # take first entry

        # apply to remote interfaces

        for edge in coll_dom.edges():
            edge.dst_int.subnet = coll_dom.subnet

    # also need to form aggregated IP blocks (used for e.g. routing prefix
    # advertisement)
    # import autonetkit
    # autonetkit.update_vis(anm)
    if len(mismatched_interfaces):
        log.warning("IPv4 Infrastructure IPs %s are not in global "
                    "loopback allocation block %s"
                    % (sorted(mismatched_interfaces), global_infra_block))

    infra_blocks = {}
    for (asn, devices) in g_ipv4.groupby('asn').items():
        broadcast_domains = [d for d in devices if d.broadcast_domain]
        subnets = [cd.subnet for cd in broadcast_domains
                   if cd.subnet is not None]  # only if subnet is set
        infra_blocks[asn] = netaddr.cidr_merge(subnets)

    # formatted = {key: [str(v) for v in val] for key, val in infra_blocks.items()}
    # log.info("Found infrastructure IP blocks %s", formatted)
    g_ipv4.data.infra_blocks = infra_blocks


#@call_log
def manual_ipv4_loopback_allocation(anm):
    """Applies manual IPv4 allocation"""

    import netaddr
    g_ipv4 = anm['ipv4']
    g_in = anm['input']

    for l3_device in g_ipv4.l3devices():
        try:
            l3_device.loopback = IPAddress(l3_device['input'].loopback_v4)
        except netaddr.AddrFormatError:
            log.debug("Unable to parse IP address %s on %s",
                      l3_device['input'].loopback_v6, l3_device)

    try:
        loopback_subnet = g_in.data.ipv4_loopback_subnet
        loopback_prefix = g_in.data.ipv4_loopback_prefix
        loopback_block = sn_preflen_to_network(loopback_subnet,
                                               loopback_prefix)
    except Exception, e:
        log.info("Unable to parse specified ipv4 loopback subnets %s/%s")
    else:
        mismatched_nodes = [n for n in g_ipv4.l3devices()
                            if n.loopback and n.loopback not in loopback_block]
        if len(mismatched_nodes):
            log.warning("IPv4 loopbacks set on nodes %s are not in global "
                        "loopback allocation block %s"
                        % (sorted(mismatched_nodes), loopback_block))

    # mismatch = [n for n in g_ipv4.l3devices() if n.loopback not in

    # also need to form aggregated IP blocks (used for e.g. routing prefix
    # advertisement)

    loopback_blocks = {}
    for (asn, devices) in g_ipv4.groupby('asn').items():
        routers = [d for d in devices if d.is_router()]
        loopbacks = [r.loopback for r in routers]
        loopback_blocks[asn] = netaddr.cidr_merge(loopbacks)

    g_ipv4.data.loopback_blocks = loopback_blocks
    # formatted = {key: [str(v) for v in val] for key, val in loopback_blocks.items()}
    #log.info("Found loopback IP blocks %s", formatted)


#@call_log
def build_ipv4(anm, infrastructure=True):
    """Builds IPv4 graph"""

    import autonetkit.plugins.ipv4 as ipv4
    import netaddr
    g_ipv4 = anm.add_overlay('ipv4')
    g_ip = anm['ip']
    g_in = anm['input']
    # retain if collision domain or not
    g_ipv4.add_nodes_from(g_ip, retain=['label', 'allocate',
                                        'broadcast_domain'])

    # Copy ASN attribute chosen for collision domains (used in alloc algorithm)

    ank_utils.copy_attr_from(
        g_ip, g_ipv4, 'asn', nbunch=g_ipv4.nodes('broadcast_domain'))
    # work around until fall-through implemented
    vswitches = [n for n in g_ip.nodes()
                 if n['layer2'].device_type == "switch"
                 and n['layer2'].device_subtype == "virtual"]
    ank_utils.copy_attr_from(g_ip, g_ipv4, 'asn', nbunch=vswitches)
    g_ipv4.add_edges_from(g_ip.edges())

    # check if ip ranges have been specified on g_in

    (infra_block, loopback_block, vrf_loopback_block) = \
        extract_ipv4_blocks(anm)

# TODO: don't present if using manual allocation
    if any(i for n in g_ip.nodes() for i in
           n.loopback_interfaces() if not i.is_loopback_zero):
        block_message = "IPv4 Secondary Loopbacks: %s" % vrf_loopback_block
        log.info(block_message)

    # See if IP addresses specified on each interface

    # do we need this still? in ANM? - differnt because input graph.... but
    # can map back to  self overlay first then phy???
    l3_devices = [d for d in g_in if d.device_type in ('router', 'server')]

    # TODO: need to account for devices whose interfaces are in only e.g. vpns

    manual_alloc_devices = set()
    for device in l3_devices:
        physical_interfaces = list(device.physical_interfaces())
        allocated = list(
            interface.ipv4_address for interface in physical_interfaces
            if interface.is_bound and interface['ipv4'].allocate is not False
            and interface['ipv4'].is_bound)
        if all(interface.ipv4_address for interface in
               physical_interfaces if interface.is_bound
               and interface['ip'].allocate is not False
               and interface['ip'].is_bound):
            # add as a manual allocated device
            manual_alloc_devices.add(device)

    if manual_alloc_devices == set(l3_devices):
        manual_alloc_ipv4_infrastructure = True
    else:
        log.info("Allocating from IPv4 infrastructure block: %s" % infra_block)
        manual_alloc_ipv4_infrastructure = False
        # warn if any set
        allocated = []
        unallocated = []
        for node in l3_devices:
            # TODO: make these inverse sets
            allocated += sorted([i for i in node.physical_interfaces()
                                 if i.is_bound and i.ipv4_address])
            unallocated += sorted([i for i in node.physical_interfaces()
                                   if i.is_bound and not i.ipv4_address
                                   and i['ipv4'].is_bound])

        # TODO: what if IP is set but not a prefix?
        if len(allocated):
            # TODO: if set is > 50% of nodes then list those that are NOT set
            log.warning(
                "Using automatic IPv4 interface allocation. IPv4 interface addresses specified on interfaces %s will be ignored." % allocated)

    # TODO: need to set allocate_ipv4 by default in the readers

    if manual_alloc_ipv4_infrastructure:
        manual_ipv4_infrastructure_allocation(anm)
    else:
        ipv4.allocate_infra(g_ipv4, infra_block)

    if g_in.data.alloc_ipv4_loopbacks is False:
        manual_ipv4_loopback_allocation(anm)
    else:
        log.info("Allocating from IPv4 loopback block: %s" % loopback_block)
        # Check if some nodes are allocated
        allocated = sorted([n for n in g_ip if n['input'].loopback_v4])
        unallocated = sorted([n for n in g_ip if not n['input'].loopback_v4])
        if len(allocated):
            log.warning(
                "Using automatic IPv4 loopback allocation. IPv4 loopback addresses specified on nodes %s will be ignored." % allocated)
            # TODO: if set is > 50% of nodes then list those that are NOT set
        ipv4.allocate_loopbacks(g_ipv4, loopback_block)

    # TODO: need to also support secondary_loopbacks for IPv6
    # TODO: only call if secondaries are set

    ipv4.allocate_secondary_loopbacks(g_ipv4, vrf_loopback_block)

    # TODO: replace this with direct allocation to interfaces in ip alloc plugin
    # TODO: add option for nonzero interfaces on node - ie
    # node.secondary_loopbacks
    for node in g_ipv4:
        node.static_routes = []

    for node in g_ipv4.routers():
        node.loopback_zero.ip_address = node.loopback
        node.loopback_zero.subnet = netaddr.IPNetwork("%s/32" % node.loopback)
        for interface in node.loopback_interfaces():
            if not interface.is_loopback_zero:
                # TODO: fix this inconsistency elsewhere
                interface.ip_address = interface.loopback



import functools
import logging
import string

import autonetkit.log as log
from autonetkit.log import CustomAdapter
from autonetkit.nidb.config_stanza import ConfigStanza
from autonetkit.nidb.interface import DmInterface
import autonetkit.log as log


@functools.total_ordering
class DmNode(object):

    """API to access overlay graph node in network"""

    def __init__(self, nidb, node_id):
        # Set using this method to bypass __setattr__
        object.__setattr__(self, 'nidb', nidb)
        object.__setattr__(self, 'node_id', node_id)
        #logger = logging.getLogger("ANK")
        # TODO: also pass the node object to the logger for building custom output lists
        # ie create a special handler that just outputs the specific node/link/interface errors
        #logstring = "Node: %s" % str(self)
        #logger = CustomAdapter(logger, {'item': logstring})
        logger = log
        object.__setattr__(self, 'log', logger)

    # TODO: make a json objct that returns keys that aren't logs, etc - filter
    # out

    def __repr__(self):
        return self._node_data['label']

    # TODO: add a dump method - needed with str()?

    def add_stanza(self, name, **kwargs):
        # TODO: decide if want shortcut for *args to set to True
        if self.get(name):
            value = self.get(name)
            if isinstance(value, ConfigStanza):
                # Don't recreate
                self.log.debug("Stanza %s already exists" % name)
                return value
            else:
                # TODO: remove? - as shouldn't reach here now? GH-186
                log.warning(
                    "Creating stanza: %s already set as %s for %s" % (name, type(value), self))

        stanza = ConfigStanza(**kwargs)
        self.__setattr__(name, stanza)
        return stanza

    def __hash__(self):
        return hash(self.node_id)

    def __eq__(self, other):
        try:
            return self.node_id == other.node_id
        except AttributeError:
            # TODO: check why comparing against strings - if in overlay
            # graph...
            return self.node_id == other

    def interface(self, key):
        # TODO: also need to allow access interface for nidb and search on
        # (node, interface id) tuple
        try:
            interface_id = key.interface_id  # eg extract from interface
        except AttributeError:
            interface_id = key  # eg string

        return DmInterface(self.nidb, self.node_id, interface_id)

    @property
    def _ports(self):
        """Returns underlying interface dict"""
        try:
            return self._graph.node[self.node_id]["_ports"]
        except KeyError:
            log.debug("No interfaces initialised for %s" % self)
            return

    @property
    def _next_int_id(self):
        """"""
# returns next free interface ID
        import itertools
        for int_id in itertools.count(1):  # start at 1 as 0 is loopback
            if int_id not in self._ports:
                return int_id

    def add_interface(self, description=None, category="physical", *args,  **kwargs):
        """Public function to add interface"""
        data = dict(kwargs)
        interface_id = self._next_int_id
        data['category'] = category  # store category on node
        data['description'] = description
        self._ports[interface_id] = data

        return DmInterface(self.nidb, self.node_id, interface_id)

    @property
    def _interface_ids(self):
        return self._ports.keys()

    @property
    def interfaces(self):
        # TODO: make not a property
        """Called by templates, sorts by ID"""
        int_list = self.get_interfaces()

        # Put loopbacks before physical interfaces
        type_index = {"loopback": 0, "physical": 1}
        # TODO: extend this based on medium category, etc

        int_list = sorted(int_list, key=lambda x: x.id)
        int_list = sorted(int_list, key=lambda x: type_index[x.category])
        return int_list

    def physical_interfaces(self):
        return self.get_interfaces(category="physical")

    def loopback_interfaces(self):
        return self.get_interfaces(category="loopback")

    def get_interfaces(self, *args, **kwargs):
        """Public function to view interfaces

        Temporary function name until Compiler/DeviceModel/Templates
        move to using "proper" interfaces"""
        def filter_func(interface):
            """Filter based on args and kwargs"""
            return (
                all(getattr(interface, key) for key in args) and
                all(getattr(
                    interface, key) == val for key, val in kwargs.items())
            )

        all_interfaces = iter(DmInterface(self.nidb,
                                          self.node_id, interface_id)
                              for interface_id in self._interface_ids)
        retval = (i for i in all_interfaces if filter_func(i))
        return retval

    @property
    def loopback_zero(self):
        return (i for i in self.interfaces if i.is_loopback_zero).next()

    @property
    def raw_interfaces(self):
        """Direct access to the interfaces dictionary, used by ANK modules"""
        return self._ports

    @raw_interfaces.setter
    def raw_interfaces(self, value):
        self._ports = value

    @property
    def _graph(self):
        return self.nidb.raw_graph()

    def degree(self):
        return self._graph.degree(self.node_id)

    def neighbors(self):
        return iter(DmNode(self.nidb, node)
                    for node in self._graph.neighbors(self.node_id))

    def __setstate__(self, state):
        (nidb, node_id) = state
        object.__setattr__(self, 'nidb', nidb)
        object.__setattr__(self, 'node_id', node_id)

    def __lt__(self, other):
        # TODO: use human sort from StackOverflow

        # sort on label if available
        if self.label is not None:
            self_node_id = self.label
        else:
            self_node_id = self.node_id

        if other.label is not None:
            other_node_id = other.label
        else:
            other_node_id = other_node_id

        try:
            self_node_string = [x for x in self_node_id if x
                                not in string.digits]
            other_node_string = [x for x in self_node_id if x
                                 not in string.digits]
        except TypeError:

            # e.g. non-iterable type, such as an int node_id

            pass
        else:
            if self_node_string == other_node_string:
                self_node_id = """""".join([x for x in self_node_id
                                            if x in string.digits])
                other_node_id = """""".join([x for x in other_node_id
                                             if x in string.digits])
                try:
                    self_node_id = int(self_node_id)
                except ValueError:
                    pass  # not a number
                try:
                    other_node_id = int(other_node_id)
                except ValueError:
                    pass  # not a number

        return (self.asn, self_node_id) < (other.asn, other_node_id)

    @property
    def _node_data(self):
        return self.nidb.raw_graph().node[self.node_id]

    def dump(self):
        # return str(self._node_data)
        import pprint
        pprint.pprint(self._node_data)

    def __nonzero__(self):
        return self.node_id in self.nidb.raw_graph()

    def is_router(self):
        return self.device_type == "router"

    def is_device_type(self, device_type):
        """Generic user-defined cross-overlay search for device_type for consistency with ANM"""
        return self.device_type == device_type

    def is_switch(self):
        return self.device_type == "switch"

    def is_server(self):
        return self.device_type == "server"

    def is_l3device(self):
        """Layer 3 devices: router, server, cloud, host
        ie not switch
        """
        # TODO: need to check for cloud, host
        return self.is_router() or self.is_server()

    def edges(self, *args, **kwargs):
        # TODO: want to add filter for *args and **kwargs here too
        return self.nidb.edges(self, *args, **kwargs)

    @property
    def id(self):
        return self.node_id

    @property
    def label(self):
        return self.__repr__()

    def get(self, key):
        return getattr(self, key)

    def __getattr__(self, key):
        """Returns edge property"""
        data = self._node_data.get(key)

        # TODO: remove once deprecated DmNode_category
        if isinstance(data, ConfigStanza):
            return data

        return data

    def __setattr__(self, key, val):
        """Sets edge property"""
        self._node_data[key] = val
        # return DmNode_category(self.nidb, self.node_id, key)

    def __iter__(self):
        return iter(self._node_data)

    def set(self, key, val):
        """For consistency, node.set(key, value) is neater
        than setattr(node, key, value)"""
        return self.__setattr__(key, val)



import ConfigParser
import os
import os.path

import pkg_resources
import validate
from autonetkit.exception import AutoNetkitException
from configobj import ConfigObj, flatten_errors

validator = validate.Validator()

# from http://stackoverflow.com/questions/4028904
ank_user_dir = os.path.join(os.path.expanduser("~"),  ".autonetkit")

def load_config():
    retval = ConfigParser.RawConfigParser()
    spec_file = pkg_resources.resource_filename(__name__,"/config/configspec.cfg")
    retval = ConfigObj(configspec=spec_file, encoding='UTF8')
# User's ANK retval
    user_config_file = os.path.join(ank_user_dir, "autonetkit.cfg")
    retval.merge(ConfigObj(user_config_file))
# ANK retval in current directory
    retval.merge(ConfigObj("autonetkit.cfg"))
# ANK retval specified by environment variable
    try:
        ankcfg = os.environ['AUTONETKIT_CFG']
        retval.merge(ConfigObj(ankcfg))
    except KeyError:
        pass

    results = retval.validate(validator)
    if results != True:
        for (section_list, key, _) in flatten_errors(retval, results):
            if key is not None:
                print "Error loading configuration file:"
                print 'Invalid key "%s" in section "%s"' % (key, ', '.join(section_list))
                raise AutoNetkitException
            else:
# ignore missing sections - use defaults
                #print 'The following section was missing:%s ' % ', '.join(section_list)
                pass
    return retval

#NOTE: this only gets loaded once package-wide if imported as import autonetkit.config
settings = load_config()



import autonetkit.log as log
import autonetkit.ank as ank_utils

from autonetkit.ank_utils import call_log

# TODO: extract the repeated code and use the layer2  and layer3 graphs


def build_igp(anm):
    build_ospf(anm)
    build_eigrp(anm)
    build_isis(anm)

    # Build a protocol summary graph
    g_igp = anm.add_overlay("igp")
    igp_protocols = ["ospf", "eigrp", "isis"]
    for protocol in igp_protocols:
        g_protocol = anm[protocol]
        g_igp.add_nodes_from(g_protocol, igp=protocol)
        g_igp.add_edges_from(g_protocol.edges(), igp=protocol)

#@call_log


def build_ospf(anm):
    """
    Build OSPF graph.

    Allowable area combinations:
    0 -> 0
    0 -> x (x!= 0)
    x -> 0 (x!= 0)
    x -> x (x != 0)

    Not-allowed:
    x -> x (x != y != 0)

    """
    import netaddr
    g_in = anm['input']
    g_l3 = anm['layer3']
    g_phy = anm['phy']
    # add regardless, so allows quick check of node in anm['ospf'] in compilers
    g_ospf = anm.add_overlay("ospf")
    if not anm['phy'].data.enable_routing:
        g_ospf.log.info("Routing disabled, not configuring OSPF")
        return

    if not any(n.igp == "ospf" for n in g_phy):
        g_ospf.log.debug("No OSPF nodes")
        return

    ospf_nodes = [n for n in g_l3 if n['phy'].igp == "ospf"]
    g_ospf.add_nodes_from(ospf_nodes)
    g_ospf.add_edges_from(g_l3.edges(), warn=False)
    ank_utils.copy_int_attr_from(g_l3, g_ospf, "multipoint")

    # TODO: work out why this doesnt work
    #ank_utils.copy_int_attr_from(g_in, g_ospf, "ospf_cost", dst_attr="cost",  type=int, default = 1)
    for node in g_ospf:
        for interface in node.physical_interfaces():
            interface.cost = 1

    ank_utils.copy_attr_from(g_in, g_ospf, "ospf_area", dst_attr="area")
    #ank_utils.copy_edge_attr_from(g_in, g_ospf, "ospf_cost", dst_attr="cost",  type=int, default = 1)
    ank_utils.copy_attr_from(
        g_in, g_ospf, "custom_config_ospf", dst_attr="custom_config")

    g_ospf.remove_edges_from([link for link in g_ospf.edges(
    ) if link.src.asn != link.dst.asn])  # remove inter-AS links

    area_zero_ip = netaddr.IPAddress("0.0.0.0")
    area_zero_int = 0
    area_zero_ids = {area_zero_ip, area_zero_int}
    default_area = area_zero_int
    if any(router.area == "0.0.0.0" for router in g_ospf):
        # string comparison as hasn't yet been cast to IPAddress
        default_area = area_zero_ip

    for router in g_ospf:
        if not router.area or router.area == "None":
            router.area = default_area
            # check if 0.0.0.0 used anywhere, if so then use 0.0.0.0 as format
        else:
            try:
                router.area = int(router.area)
            except ValueError:
                try:
                    router.area = netaddr.IPAddress(router.area)
                except netaddr.core.AddrFormatError:
                    router.log.warning("Invalid OSPF area %s. Using default"
                                       " of %s" % (router.area, default_area))
                    router.area = default_area

    # TODO: use interfaces throughout, rather than edges
    for router in g_ospf:
        # and set area on interface
        for edge in router.edges():
            if edge.area:
                continue  # allocated (from other "direction", as undirected)
            if router.area == edge.dst.area:
                edge.area = router.area  # intra-area
                continue

            if router.area in area_zero_ids or edge.dst.area in area_zero_ids:
                # backbone to other area
                if router.area in area_zero_ids:
                    # router in backbone, use other area
                    edge.area = edge.dst.area
                else:
                    # router not in backbone, use its area
                    edge.area = router.area

    for router in g_ospf:
        areas = {edge.area for edge in router.edges()}
        router.areas = list(areas)  # edges router participates in

        if len(areas) in area_zero_ids:
            router.type = "backbone"  # no ospf edges (eg single node in AS)
        elif len(areas) == 1:
            # single area: either backbone (all 0) or internal (all nonzero)
            if len(areas & area_zero_ids):
                # intersection has at least one element -> router has area zero
                router.type = "backbone"
            else:
                router.type = "internal"

        else:
            # multiple areas
            if len(areas & area_zero_ids):
                # intersection has at least one element -> router has area zero
                router.type = "backbone ABR"
            elif router.area in area_zero_ids:
                router.log.debug(
                    "Router belongs to area %s but has no area zero interfaces",
                    router.area)
                router.type = "backbone ABR"
            else:
                router.log.warning(
                    "spans multiple areas but is not a member of area 0")
                router.type = "INVALID"

    if (any(area_zero_int in router.areas for router in g_ospf) and
            any(area_zero_ip in router.areas for router in g_ospf)):
        router.log.warning("Using both area 0 and area 0.0.0.0")

    for link in g_ospf.edges():
        if not link.cost:
            link.cost = 1

    # map areas and costs onto interfaces
    # TODO: later map them directly rather than with edges - part of
    # the transition
    for edge in g_ospf.edges():
        for interface in edge.interfaces():
            interface.cost = edge.cost
            interface.area = edge.area
            interface.multipoint = edge.multipoint

    for router in g_ospf:
        router.loopback_zero.area = router.area
        router.loopback_zero.cost = 0
        router.process_id = router.asn

#@call_log


def ip_to_net_ent_title_ios(ip_addr):
    """ Converts an IP address into an OSI Network Entity Title
    suitable for use in IS-IS on IOS.

    >>> from netaddr import IPAddress
    >>> ip_to_net_ent_title_ios(IPAddress("192.168.19.1"))
    '49.1921.6801.9001.00'
    """
    try:
        ip_words = ip_addr.words
    except AttributeError:
        import netaddr  # try to cast to IP Address
        ip_addr = netaddr.IPAddress(ip_addr)
        ip_words = ip_addr.words

    log.debug("Converting IP to OSI ENT format")
    area_id = "49"
    ip_octets = "".join("%03d" % int(
        octet) for octet in ip_words)  # single string, padded if needed
    return ".".join([area_id, ip_octets[0:4], ip_octets[4:8], ip_octets[8:12],
                     "00"])

#@call_log


def build_eigrp(anm):
    """Build eigrp overlay"""
    g_in = anm['input']
    # add regardless, so allows quick check of node in anm['isis'] in compilers
    g_l3 = anm['layer3']
    g_eigrp = anm.add_overlay("eigrp")
    g_phy = anm['phy']

    if not anm['phy'].data.enable_routing:
        g_eigrp.log.info("Routing disabled, not configuring EIGRP")
        return

    if not any(n.igp == "eigrp" for n in g_phy):
        log.debug("No EIGRP nodes")
        return
    eigrp_nodes = [n for n in g_l3 if n['phy'].igp == "eigrp"]
    g_eigrp.add_nodes_from(eigrp_nodes)
    g_eigrp.add_edges_from(g_l3.edges(), warn=False)
    ank_utils.copy_int_attr_from(g_l3, g_eigrp, "multipoint")

    ank_utils.copy_attr_from(
        g_in, g_eigrp, "custom_config_eigrp", dst_attr="custom_config")

# Merge and explode switches
    ank_utils.aggregate_nodes(g_eigrp, g_eigrp.switches())
    exploded_edges = ank_utils.explode_nodes(g_eigrp,
                                             g_eigrp.switches())
    for edge in exploded_edges:
        edge.multipoint = True

    g_eigrp.remove_edges_from(
        [link for link in g_eigrp.edges() if link.src.asn != link.dst.asn])

    for node in g_eigrp:
        node.process_id = node.asn

    for link in g_eigrp.edges():
        link.metric = 1  # default

    for edge in g_eigrp.edges():
        for interface in edge.interfaces():
            interface.metric = edge.metric
            interface.multipoint = edge.multipoint

#@call_log


def build_network_entity_title(anm):
    g_isis = anm['isis']
    g_ipv4 = anm['ipv4']
    for node in g_isis.routers():
        ip_node = g_ipv4.node(node)
        node.net = ip_to_net_ent_title_ios(ip_node.loopback)


def build_isis(anm):
    """Build isis overlay"""
    g_in = anm['input']
    # add regardless, so allows quick check of node in anm['isis'] in compilers
    g_l3 = anm['layer3']
    g_phy = anm['phy']
    g_isis = anm.add_overlay("isis")

    if not anm['phy'].data.enable_routing:
        g_isis.log.info("Routing disabled, not configuring ISIS")
        return

    if not any(n.igp == "isis" for n in g_phy):
        g_isis.log.debug("No ISIS nodes")
        return

    isis_nodes = [n for n in g_l3 if n['phy'].igp == "isis"]
    g_isis.add_nodes_from(isis_nodes)
    g_isis.add_edges_from(g_l3.edges(), warn=False)
    ank_utils.copy_int_attr_from(g_l3, g_isis, "multipoint")

    ank_utils.copy_attr_from(
        g_in, g_isis, "custom_config_isis", dst_attr="custom_config")

    g_isis.remove_edges_from(
        [link for link in g_isis.edges() if link.src.asn != link.dst.asn])

    build_network_entity_title(anm)

    for node in g_isis.routers():
        node.process_id = node.asn

    for link in g_isis.edges():
        link.metric = 1  # default

    for edge in g_isis.edges():
        for interface in edge.interfaces():
            interface.metric = edge.metric
            interface.multipoint = edge.multipoint



class AutoNetkitException(Exception):

    """Base class for AutoNetkit Exceptions"""


class AnkIncorrectFileFormat(AutoNetkitException):

    """Wrong file format"""


class OverlayNotFound(AutoNetkitException):

    def __init__(self, errors):
        self.Errors = errors

    def __str__(self):
        return 'Overlay %s not found' % self.Errors



import autonetkit.log as log

from autonetkit.nidb.interface import DmInterface
from autonetkit.nidb.edge import DmEdge
from autonetkit.nidb.node import DmNode
from autonetkit import ank_json

# TODO: add to doc we don't llow bidirectional nidb


class DmBase(object):
    # TODO: inherit common methods from same base as overlay

    def __init__(self):
        # TODO: make optional for restore serialized file on init
        self._graph = None
        pass

    def __getstate__(self):
        return self._graph

    def __setstate__(self, state):
        self._graph = state

    def __repr__(self):
        return "nidb"

    def is_multigraph(self):
        return self._graph.is_multigraph()

    # Model-level functions

    def save(self, timestamp=True, use_gzip=True):
        import os
        import gzip
        archive_dir = os.path.join("versions", "nidb")
        if not os.path.isdir(archive_dir):
            os.makedirs(archive_dir)

        data = ank_json.ank_json_dumps(self._graph)
        if timestamp:
            json_file = "nidb_%s.json.gz" % self.timestamp
        else:
            json_file = "nidb.json"
        json_path = os.path.join(archive_dir, json_file)
        log.debug("Saving to %s" % json_path)
        if use_gzip:
            with gzip.open(json_path, "wb") as json_fh:
                json_fh.write(data)
        else:
            with open(json_path, "wb") as json_fh:
                json_fh.write(data)

    def interface(self, interface):
        return DmInterface(self,
                           interface.node_id, interface.interface_id)

    def restore_latest(self, directory=None):
        import os
        import glob
        if not directory:
            # TODO: make directory loaded from config
            directory = os.path.join("versions", "nidb")

        glob_dir = os.path.join(directory, "*.json.gz")
        pickle_files = glob.glob(glob_dir)
        pickle_files = sorted(pickle_files)
        try:
            latest_file = pickle_files[-1]
        except IndexError:
            # No files loaded
            log.warning(
                "No previous DeviceModel saved. Please compile new DeviceModel")
            return
        self.restore(latest_file)
        ank_json.rebind_nidb_interfaces(self)

    def restore(self, pickle_file):
        import gzip
        log.debug("Restoring %s" % pickle_file)
        with gzip.open(pickle_file, "r") as fh:
            #data = json.load(fh)
            data = fh.read()
            self._graph = ank_json.ank_json_loads(data)

        ank_json.rebind_nidb_interfaces(self)

    @property
    def name(self):
        return self.__repr__()

    def raw_graph(self):
        """Returns the underlying NetworkX graph"""
        return self._graph

    def copy_graphics(self, network_model):
        """Transfers graphics data from anm to nidb"""
        for node in self:
            node.add_stanza("graphics")
            #TODO: deprecate g_graphics overlay, store on phy
            if (network_model.has_overlay("graphics")
                and node in network_model['graphics']):
                source_node = network_model['graphics'].node(node)
            else:
                source_node = network_model['phy'].node(node)
            node.graphics.x = source_node.x
            node.graphics.y = source_node.y
            node.graphics.device_type = source_node.device_type
            node.graphics.device_subtype = source_node.device_subtype
            node.device_type = source_node.device_type
            node.device_subtype = source_node.device_subtype

    def __len__(self):
        return len(self._graph)

    @property
    def data(self):
        from autonetkit.nidb.device_model import DmGraphData
        return DmGraphData(self)

    # Nodes

    def __iter__(self):
        return iter(DmNode(self, node)
                    for node in self._graph)

    def node(self, key):
        """Returns node based on name
        This is currently O(N). Could use a lookup table"""
        try:
            if key.node_id in self._graph:
                return DmNode(self, key.node_id)
        except AttributeError:
            # doesn't have node_id, likely a label string, search on label
            for node in self:
                if str(node) == key:
                    return node
                elif node.id == key:
                    # label could be "a b" -> "a_b" (ie folder safe, etc)
                    # TODO: need to fix this discrepancy
                    return node
            print "Unable to find node", key, "in", self
            return None

    def update(self, nbunch, **kwargs):
        for node in nbunch:
            for (_, key), value in kwargs.items():
                node.category.set(key, value)

    def nodes(self, *args, **kwargs):
        result = self.__iter__()
        if len(args) or len(kwargs):
            result = self.filter(result, *args, **kwargs)
        return result

    def routers(self, *args, **kwargs):
        """Shortcut for nodes(), sets device_type to be router"""

        result = self.nodes(*args, **kwargs)
        return [r for r in result if r.is_router()]

    def switches(self, *args, **kwargs):
        """Shortcut for nodes(), sets device_type to be switch"""

        result = self.nodes(*args, **kwargs)
        return [r for r in result if r.is_switch()]

    def servers(self, *args, **kwargs):
        """Shortcut for nodes(), sets device_type to be server"""

        result = self.nodes(*args, **kwargs)
        return [r for r in result if r.is_server()]

    def l3devices(self, *args, **kwargs):
        """Shortcut for nodes(), sets device_type to be server"""
        result = self.nodes(*args, **kwargs)
        return [r for r in result if r.is_l3device()]

    def filter(self, nbunch=None, *args, **kwargs):
        # TODO: also allow nbunch to be passed in to subfilter on...?
        """TODO: expand this to allow args also, ie to test if value evaluates to True"""
        # need to allow filter_func to access these args
        if not nbunch:
            nbunch = self.nodes()

        def filter_func(node):
            return (
                all(getattr(node, key) for key in args) and
                all(getattr(node, key) == val for key, val in kwargs.items())
            )

        return (n for n in nbunch if filter_func(n))

    def add_nodes_from(self, nbunch, retain=None, **kwargs):
        if retain is None:
            retain = []
        try:
            retain.lower()
            retain = [retain]  # was a string, put into list
        except AttributeError:
            pass  # already a list

        nbunch = list(nbunch)
        nodes_to_add = nbunch  # retain for interface copying

        if len(retain):
            add_nodes = []
            for n in nbunch:
                data = dict((key, n.get(key)) for key in retain)
                add_nodes.append((n.node_id, data))
            nbunch = add_nodes
        else:
            log.warning(
                "Cannot add node ids directly to DeviceModel: must add overlay nodes")
        self._graph.add_nodes_from(nbunch, **kwargs)

        for node in nodes_to_add:
            # TODO: add an interface_retain for attributes also
            int_dict = {i.interface_id: {'category': i.category,
                                         'description': i.description,
                                         'layer': i.overlay_id} for i in node.interfaces()}
            int_dict = {i.interface_id: {'category': i.category,
                                         'description': i.description,
                                         } for i in node.interfaces()}
            self._graph.node[node.node_id]["_ports"] = int_dict

    # Edges

    def edges(self, nbunch=None, *args, **kwargs):
        # nbunch may be single node
        # TODO: Apply edge filters
        if nbunch:
            try:
                nbunch = nbunch.node_id
            except AttributeError:
                # only store the id in overlay
                nbunch = (n.node_id for n in nbunch)

        def filter_func(edge):
            return (
                all(getattr(edge, key) for key in args) and
                all(getattr(edge, key) == val for key, val in kwargs.items())
            )

        # TODO: See if more efficient way to access underlying data structure
        # rather than create overlay to throw away
        if self.is_multigraph():
            valid_edges = list((src, dst, key) for (src, dst, key) in
                               self._graph.edges(nbunch, keys=True))
        else:
            default_key = 0
            valid_edges = list((src, dst, default_key) for (src, dst) in
                               self._graph.edges(nbunch))

        all_edges = [DmEdge(self, src, dst, key)
                     for src, dst, key in valid_edges]
        return (edge for edge in all_edges if filter_func(edge))

    def edge(self, edge_to_find, key=0):
        """returns edge in this graph with same src and dst"""
        # TODO: check if this even needed - will be if searching nidb specifically
        # but that's so rare (that's a design stage if anywhere)
        src_id = edge_to_find.src
        dst_id = edge_to_find.dst

        if self.is_multigraph():
            for (src, dst, rkey) in self._graph.edges(src_id, keys=True):
                if dst == dst_id and rkey == search_key:
                    return DmEdge(self._anm, src, dst, search_key)

        for (src, dst) in self._graph.edges(src_id):
            if dst == dst_id:
                return DmEdge(self._anm, src, dst)

    def add_edge(self, src, dst, retain=None, **kwargs):
        # TODO: support multigraph
        if retain is None:
            retain = []
        self.add_edges_from([(src, dst)], retain, **kwargs)

    def add_edges_from(self, ebunch, retain=None, **kwargs):
        """Used to copy edges from ANM -> NIDB
        Note: won't support (yet) copying from one NIDB to another
        #TODO: allow copying from one NIDB to another
        (check for DmNode as well as NmNode)

        To keep congruency, only allow copying edges from ANM
        can't add NIDB edges directly (node, node) oor (port, port)
        workflow: if need this, create a new overlay and copy from there
        """
        from autonetkit.anm import NmEdge
        if not retain:
            retain = []
        try:
            retain.lower()
            retain = [retain]  # was a string, put into list
        except AttributeError:
            pass  # already a list

        # TODO: this needs to support parallel links
        for in_edge in ebunch:
            """Edge could be one of:
            - NmEdge - copied into be returned as a DmEdge
            """
            # This is less efficient than nx add_edges_from, but cleaner logic
            # TODO: could put the interface data into retain?
            data = {'_ports': {}}  # to retain
            ekey = 0

            # convert input to a NmEdge
            if isinstance(in_edge, NmEdge):
                edge = in_edge  # simple case
                ekey = edge.ekey
                src = edge.src.node_id
                dst = edge.dst.node_id

                # and copy retain data
                retain.append('_ports')
                # TODO: explicity copy ports as raw_interfaces?
                data = dict((key, edge.get(key)) for key in retain)

                # this is the only case where copy across data
                # but want to copy attributes for all cases

            # TODO: add check that edge.src and edge.dst exist
            if not(src in self and dst in self):
                log.warning("Not adding edge, %s to %s, "
                            "src and/or dst not in overlay %s" % (src, dst, self))
                continue

            # TODO: warn if not multigraph and edge already exists - don't
            # add/clobber
            data.update(**kwargs)

            if self.is_multigraph():
                self._graph.add_edge(src, dst, key=ekey,
                                     attr_dict=dict(data))
            else:
                self._graph.add_edge(src, dst, attr_dict=dict(data))



"""Console script entry point for AutoNetkit"""

import os
import sys
import time
import traceback
from datetime import datetime

import autonetkit.config as config
import autonetkit.log as log
import autonetkit.workflow as workflow
import pkg_resources

try:
    ANK_VERSION = pkg_resources.get_distribution("autonetkit").version
except pkg_resources.DistributionNotFound:
    ANK_VERSION = "dev"

def parse_options(argument_string=None):
    """Parse user-provided options"""
    import argparse
    usage = "autonetkit -f input.graphml"
    version = "%(prog)s " + str(ANK_VERSION)
    parser = argparse.ArgumentParser(description=usage, version=version)

    input_group = parser.add_mutually_exclusive_group()
    input_group.add_argument(
        '--file', '-f', default=None, help="Load topology from FILE")
    input_group.add_argument('--stdin', action="store_true", default=False,
                             help="Load topology from STDIN")

    parser.add_argument(
        '--monitor', '-m', action="store_true", default=False,
        help="Monitor input file for changes")
    parser.add_argument('--debug', action="store_true",
                        default=False, help="Debug mode")
    parser.add_argument('--quiet', action="store_true",
                        default=False, help="Quiet mode (only display warnings and errors)")
    parser.add_argument('--diff', action="store_true", default=False,
                        help="Diff DeviceModel")
    parser.add_argument('--no_vis', dest="visualise",
        action="store_false", default=True,
        help="Visualise output")
    parser.add_argument('--compile', action="store_true",
                        default=False, help="Compile")
    parser.add_argument(
        '--build', action="store_true", default=False, help="Build")
    parser.add_argument(
        '--render', action="store_true", default=False, help="Compile")
    parser.add_argument(
        '--validate', action="store_true", default=False, help="Validate")
    parser.add_argument('--deploy', action="store_true",
                        default=False, help="Deploy")
    parser.add_argument('--archive', action="store_true", default=False,
                        help="Archive ANM, DeviceModel, and IP allocations")
    parser.add_argument('--measure', action="store_true",
                        default=False, help="Measure")
    parser.add_argument(
        '--webserver', action="store_true", default=False, help="Webserver")
    parser.add_argument('--grid', type=int, help="Grid Size (n * n)")
    parser.add_argument(
        '--target', choices=['netkit', 'cisco'], default=None)
    parser.add_argument(
        '--vis_uuid', default=None, help="UUID for multi-user visualisation")
    if argument_string:
        arguments = parser.parse_args(argument_string.split())
    else:
        # from command line arguments
        arguments = parser.parse_args()

    return arguments


def main(options):
    settings = config.settings

    if options.vis_uuid:
        config.settings['Http Post']['uuid'] = options.vis_uuid

    try:
        # test if can import, if not present will fail and not add to template
        # path
        import autonetkit_cisco
    except ImportError:
        pass
    else:
        import autonetkit_cisco.version
        version_banner = autonetkit_cisco.version.banner()
        log.info("%s" % version_banner)

    log.info("AutoNetkit %s" % ANK_VERSION)

    if options.target == "cisco":
        # output target is Cisco
        log.info("Setting output target as Cisco")
        settings['Graphml']['Node Defaults']['platform'] = "VIRL"
        settings['Graphml']['Node Defaults']['host'] = "internal"
        settings['Graphml']['Node Defaults']['syntax'] = "ios_xr"
        settings['Compiler']['Cisco']['to memory'] = 1
        settings['General']['deploy'] = 1
        settings['Deploy Hosts']['internal'] = {'VIRL':
                                                {'deploy': 1}}

    if options.debug or settings['General']['debug']:
        # TODO: fix this
        import logging
        logger = logging.getLogger("ANK")
        logger.setLevel(logging.DEBUG)

    if options.quiet or settings['General']['quiet']:
        import logging
        logger = logging.getLogger("ANK")
        logger.setLevel(logging.WARNING)

    build_options = {
        'compile': options.compile or settings['General']['compile'],
        'render': options.render or settings['General']['render'],
        'validate': options.validate or settings['General']['validate'],
        'build': options.build or settings['General']['build'],
        'deploy': options.deploy or settings['General']['deploy'],
        'measure': options.measure or settings['General']['measure'],
        'monitor': options.monitor or settings['General']['monitor'],
        'diff': options.diff or settings['General']['diff'],
        'archive': options.archive or settings['General']['archive'],
        # use and for visualise as no_vis negates
        'visualise': options.visualise and settings['General']['visualise'],
    }

    if options.webserver:
        log.info("Webserver not yet supported, please run as seperate module")

    if options.file:
        with open(options.file, "r") as fh:
            input_string = fh.read()
        timestamp = os.stat(options.file).st_mtime
    elif options.stdin:
        input_string = sys.stdin
        now = datetime.now()
        timestamp = now.strftime("%Y%m%d_%H%M%S_%f")
    elif options.grid:
        input_string = ""
        now = datetime.now()
        timestamp = now.strftime("%Y%m%d_%H%M%S_%f")
    else:
        log.info("No input file specified. Exiting")
        return

    try:
        workflow.manage_network(input_string, timestamp,
                       grid=options.grid, **build_options)
    except Exception, err:
        log.error(
            "Error generating network configurations: %s. More information may be available in the debug log." % err)
        log.debug("Error generating network configurations", exc_info=True)
        if settings['General']['stack_trace']:
            print traceback.print_exc()
        sys.exit("Unable to build configurations.")

# TODO: work out why build_options is being clobbered for monitor mode
    build_options['monitor'] = options.monitor or settings['General'][
        'monitor']

    if build_options['monitor']:
        try:
            log.info("Monitoring for updates...")
            input_filemonitor = workflow.file_monitor(options.file)
            #build_filemonitor = file_monitor("autonetkit/build_network.py")
            while True:
                time.sleep(1)
                rebuild = False
                if input_filemonitor.next():
                    rebuild = True

                if rebuild:
                    try:
                        log.info("Input graph updated, recompiling network")
                        with open(options.file, "r") as fh:
                            input_string = fh.read()  # read updates
                        workflow.manage_network(input_string,
                                       timestamp, build_options)
                        log.info("Monitoring for updates...")
                    except Exception, e:
                        log.warning("Unable to build network %s" % e)
                        traceback.print_exc()

        except KeyboardInterrupt:
            log.info("Exiting")

def console_entry():
    """If come from console entry point"""
    args = parse_options()
    main(args)

if __name__ == "__main__":
    try:
        args = parse_options()
        main(args)
    except KeyboardInterrupt:
        pass



import autonetkit.log as log


def build_reverse_mappings_from_nidb(nidb):
    """Builds IP reverse mappings from DeviceModel"""
    rev_map = {
        "subnets": {},
        "loopbacks": {},
        "infra_interfaces": {},
    }

    for node in nidb:
        if node.broadcast_domain:
            rev_map["subnets"][str(node.ipv4_subnet)] = node
        else:
            rev_map["loopbacks"][str(node.loopback)] = node
            for interface in node.physical_interfaces():
                rev_map["infra_interfaces"][
                    str(interface.ipv4_address)] = interface

    return rev_map


def build_reverse_mappings_from_anm_input(anm):
    """Builds reverse mappings from ANM input graph,
    assumes addresses have already been allocated onto input graph,
    either externally or by previous run"""

    from collections import defaultdict
    import netaddr

    g_in = anm['input']
    rev_map = {
        "loopbacks": {},
        "infra_interfaces": {},
        "subnets": {},
    }

    subnets = defaultdict(list)

    for node in g_in:
        rev_map["loopbacks"][str(node.loopback_v4)] = node
        for interface in node.physical_interfaces():
            rev_map["infra_interfaces"][
                str(interface.ipv4_address)] = interface

            prefixlen = interface.ipv4_prefixlen
            cidr_string = "%s/%s" % (interface.ipv4_address, prefixlen)
            intermediate_subnet = netaddr.IPNetwork(cidr_string)
            subnet_cidr_string = "%s/%s" % (
                intermediate_subnet.network, prefixlen)
            subnet = netaddr.IPNetwork(subnet_cidr_string)
            subnets[subnet].append(interface)

    for subnet, interfaces in subnets.items():
        subnet_str = str(subnet)
        rev_map['subnets'][subnet_str] = "_".join(
            str(i.node) for i in interfaces)

    return rev_map


def process_textfsm(template_file, data):
    """
    TODO: make template return data other than just hops, and have reverse_map_path() handle accordingly
    """
    import textfsm
    with open(template_file, "r") as template:
        re_table = textfsm.TextFSM(template)

    data = re_table.ParseText(data)
    header = re_table.header
    return header, data


def extract_route_from_parsed_routing_table(header, routes, proto_id="Proto",
                                            network_id="Network", via_id="Via"):
    network_index = header.index(network_id)
    proto_index = header.index(proto_id)
    via_index = header.index(via_id)
    return [(item[proto_index], item[network_index], item[via_index]) for item in routes]


def extract_path_from_parsed_traceroute(header, routes, hop_id="Hop"):
    """Returns the hop IPs from the TextFSM returned data"""
    hop_index = header.index(hop_id)
    return [item[hop_index] for item in routes]


def reverse_map_routing(rev_map, data):
    """Returns list of nodes in path
    interfaces selects whether to return only nodes, or interfaces
    e.g. eth0.r1 or just r1
    """

    # print data

    result = []
    for protocol, network, via in data:
        print "reversing", protocol, network, via
        if network in rev_map['subnets']:
            cd = rev_map['subnets'][network]
            if via is None:
                result.append((protocol, cd, None))

            if via in rev_map['infra_interfaces']:
                iface = rev_map['infra_interfaces'][via]
                print "adding", protocol, cd, iface.node
                result.append((protocol, cd, iface.node))
    return result


def reverse_map_address(rev_map, address, interfaces=False):
    if address in rev_map['infra_interfaces']:
        iface = rev_map['infra_interfaces'][address]
        if interfaces:
            return iface
        else:
            return iface.node
    elif address in rev_map['loopbacks']:
        node = rev_map['loopbacks'][address]
        return node


def extract_node_path_info(header, parsed_data, mapped_data, exclude_keys=None):
    if len(parsed_data) != len(mapped_data):
        log.warning(
            "Parsed data different length to mapped data, not extracting node data")

    if exclude_keys:
        exclude_keys = set(exclude_keys)
    else:
        exclude_keys = set()  # empty set for simpler test logic

    retval = []
    for index, hop in enumerate(mapped_data):
        node_vals = parsed_data[index]  # TextFSM output for this hop
        node_data = dict(zip(header, node_vals))

        filtered_data = {k: v for k, v in node_data.items()
                         if len(v) and k not in exclude_keys}
        filtered_data['host'] = str(hop.id)
        retval.append(filtered_data)

    return retval


def reverse_map_path(rev_map, path, interfaces=False):
    """Returns list of nodes in path
    interfaces selects whether to return only nodes, or interfaces
    e.g. eth0.r1 or just r1
    """

    result = []
    for hop in path:
        if hop in rev_map['infra_interfaces']:
            iface = rev_map['infra_interfaces'][hop]
            if interfaces:
                result.append(iface)
            else:
                result.append(iface.node)
        elif hop in rev_map['loopbacks']:
            node = rev_map['loopbacks'][hop]
            result.append(node)

    return result


def substitute_ips(data, rev_map, interfaces=False):
    import re
    re_ip_address = r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}"
    re_ip_loopback = r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/32"
    re_ip_subnet = r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/\d{1,2}"
    if not len(data):
        log.info("No data provided to IP substitution, returning")
        return data

    def replace_ip(match):
        match = match.group()
        if match in rev_map['infra_interfaces']:
            iface = rev_map['infra_interfaces'][match]
            if interfaces:
                named = "%s.%s" % (iface.id, iface.node)
                return named
            else:
                return str(iface.node)
        return match  # no match, return the raw IP

    def replace_loopbacks(match):
        match = match.group()
        # strip off the /32
        loopback_ip = match[:-3]
        if loopback_ip in rev_map['loopbacks']:
            node = rev_map['loopbacks'][loopback_ip]
            return str(node)
        return match  # no match, return the raw IP

    def replace_loopbacks_no_mask(match):
        # TODO: refactor
        match = match.group()
        # strip off the /32
        loopback_ip = match
        if loopback_ip in rev_map['loopbacks']:
            node = rev_map['loopbacks'][loopback_ip]
            return str(node)
        return match  # no match, return the raw IP

    def replace_subnet(match):
        match = match.group()

        if match in rev_map['subnets']:
            subnet = rev_map['subnets'][match]
            return str(subnet)
        return match  # no match, return the raw IP

    # do loopbacks first
    data = re.sub(re_ip_loopback, replace_loopbacks, data)
    data = re.sub(re_ip_address, replace_ip, data)
    # try for ip addresses in loopback
    data = re.sub(re_ip_address, replace_loopbacks_no_mask, data)
    # and for subnets ie ip/netmask
    return re.sub(re_ip_subnet, replace_subnet, data)



import autonetkit.ank as ank_utils
import autonetkit.log as log


def build_layer2(anm):
    build_layer2_base(anm)
    build_vlans(anm)
    check_layer2(anm)
    build_layer2_broadcast(anm)

def build_layer2_base(anm):
    g_l2 = anm.add_overlay('layer2')
    g_phy = anm['phy']
    g_l2.add_nodes_from(g_phy)
    g_l2.add_edges_from(g_phy.edges())
    # Don't aggregate managed switches
    unmanaged_switches = [n for n in g_l2.switches()
                          if n.device_subtype != "managed"]
    ank_utils.aggregate_nodes(g_l2, unmanaged_switches)

    try:
        from autonetkit_cisco import build_network as cisco_build_network
    except ImportError:
        pass
    else:
        cisco_build_network.post_layer2(anm)

def check_layer2(anm):
    """Sanity checks on topology"""
    from collections import defaultdict
    g_l2 = anm['layer2']

    # check for igp and ebgp on same switch
    for switch in sorted(g_l2.switches()):
        neigh_asns = defaultdict(int)
        for neigh in switch.neighbors():
            if neigh.asn is None:
                continue  # don't add if not set
            neigh_asns[neigh.asn] += 1

        # IGP if two or more neighbors share the same ASN
        is_igp = any(asns > 1 for asns in neigh_asns.values())
        # eBGP if more than one unique neigh ASN
        is_ebgp = len(neigh_asns.keys()) > 1
        if is_igp and is_ebgp:
            log.warning("Switch %s contains both IGP and eBGP neighbors",
                        switch)

    # check for multiple links from nodes to switch
    for switch in sorted(g_l2.switches()):
        for neighbor in sorted(switch.neighbors()):
            edges = g_l2.edges(switch, neighbor)
            if len(edges) > 1:
                # more than one edge between the (src, dst) pair -> parallel
                log.warning("Multiple edges (%s) between %s and device %s",
                            len(edges), switch, neighbor)

def build_layer2_broadcast(anm):
    g_l2 = anm['layer2']
    g_phy = anm['phy']
    g_graphics = anm['graphics']
    g_l2_bc = anm.add_overlay('layer2_bc')
    g_l2_bc.add_nodes_from(g_l2.l3devices())
    g_l2_bc.add_nodes_from(g_l2.switches())
    g_l2_bc.add_edges_from(g_l2.edges())

    # remove external connectors

    edges_to_split = [edge for edge in g_l2_bc.edges()
                      if edge.src.is_l3device() and edge.dst.is_l3device()]
    # TODO: debug the edges to split
    # print "edges to split", edges_to_split
    for edge in edges_to_split:
        edge.split = True  # mark as split for use in building nidb

    split_created_nodes = list(ank_utils.split(g_l2_bc, edges_to_split,
                                               retain=['split'],
                                               id_prepend='cd_'))

    # TODO: if parallel nodes, offset
    # TODO: remove graphics, assign directly
    if len(g_graphics):
        co_ords_overlay = g_graphics # source from graphics overlay
    else:
        co_ords_overlay = g_phy # source from phy overlay

    for node in split_created_nodes:
        node['graphics'].x = ank_utils.neigh_average(g_l2_bc, node, 'x',
                                                     co_ords_overlay) + 0.1

        # temporary fix for gh-90

        node['graphics'].y = ank_utils.neigh_average(g_l2_bc, node, 'y',
                                                     co_ords_overlay) + 0.1

        # temporary fix for gh-90

        asn = ank_utils.neigh_most_frequent(
            g_l2_bc, node, 'asn', g_phy)  # arbitrary choice
        node['graphics'].asn = asn
        node.asn = asn  # need to use asn in IP overlay for aggregating subnets

    # also allocate an ASN for virtual switches
    vswitches = [n for n in g_l2_bc.nodes()
                 if n['layer2'].device_type == "switch"
                 and n['layer2'].device_subtype == "virtual"]
    for node in vswitches:
        # TODO: refactor neigh_most_frequent to allow fallthrough attributes
        # asn = ank_utils.neigh_most_frequent(g_l2_bc, node, 'asn', g_l2)  #
        # arbitrary choice
        asns = [n['layer2'].asn for n in node.neighbors()]
        asns = [x for x in asns if x is not None]
        asn = ank_utils.most_frequent(asns)
        node.asn = asn  # need to use asn in IP overlay for aggregating subnets
        # also mark as broadcast domain

    from collections import defaultdict
    coincident_nodes = defaultdict(list)
    for node in split_created_nodes:
        coincident_nodes[(node['graphics'].x, node['graphics'].y)].append(node)

    coincident_nodes = {k: v for k, v in coincident_nodes.items()
                        if len(v) > 1}  # trim out single node co-ordinates
    import math
    for _, val in coincident_nodes.items():
        for index, item in enumerate(val):
            index = index + 1
            x_offset = 25 * math.floor(index / 2) * math.pow(-1, index)
            y_offset = -1 * 25 * math.floor(index / 2) * math.pow(-1, index)
            item['graphics'].x = item['graphics'].x + x_offset
            item['graphics'].y = item['graphics'].y + y_offset

    switch_nodes = g_l2_bc.switches()  # regenerate due to aggregated
    g_l2_bc.update(switch_nodes, broadcast_domain=True)

    # switches are part of collision domain
    g_l2_bc.update(split_created_nodes, broadcast_domain=True)

    # Assign collision domain to a host if all neighbours from same host

    for node in split_created_nodes:
        if ank_utils.neigh_equal(g_l2_bc, node, 'host', g_phy):
            node.host = ank_utils.neigh_attr(g_l2_bc, node, 'host',
                                             g_phy).next()  # first attribute

    # set collision domain IPs
    # TODO; work out why this throws a json exception
    #autonetkit.ank.set_node_default(g_l2_bc,  broadcast_domain=False)

    for node in g_l2_bc.nodes('broadcast_domain'):
        graphics_node = g_graphics.node(node)
        #graphics_node.device_type = 'broadcast_domain'
        if node.is_switch():
            # TODO: check not virtual
            node['phy'].broadcast_domain = True
        if not node.is_switch():
            # use node sorting, as accomodates for numeric/string names
            graphics_node.device_type = 'broadcast_domain'
            neighbors = sorted(neigh for neigh in node.neighbors())
            label = '_'.join(neigh.label for neigh in neighbors)
            cd_label = 'cd_%s' % label  # switches keep their names
            node.label = cd_label
            graphics_node.label = cd_label
            node.device_type = "broadcast_domain"
            node.label = node.id
            graphics_node.label = node.id

    for node in vswitches:
        node.broadcast_domain = True


def build_vlans(anm):
    import itertools
    from collections import defaultdict
    g_l2 = anm['layer2']
    g_vlan_trunk = anm.add_overlay('vlan_trunk')
    g_vlans = anm.add_overlay('vlans')
    managed_switches = [n for n in g_l2.switches()
                        if n.device_subtype == "managed"]

    # copy across vlans from input graph
    for router in g_l2.routers():
        for interface in router.physical_interfaces():
            interface.vlan = interface['input'].vlan
            if not interface.vlan:
                pass
                # check if connectde to a managed switch, if so, warn

    vswitch_id_counter = itertools.count(1)

    subs = ank_utils.connected_subgraphs(g_l2, managed_switches)
    for sub in subs:
        # identify the VLANs on these switches
        vlans = defaultdict(list)
        for switch in sub:
            neigh_ints = [i.neighbors()[0] for i in switch.interfaces()
                          if i.is_bound]
            router_ints = [i for i in neigh_ints if i.node.is_router()]
            for interface in router_ints:
                # store keyed by vlan id
                vlans[interface.vlan].append(interface)

        # create a virtual switch for each
        # TODO: naming: if this is the only pair then name after these, else
        # use the switch names too
        #vswitch_prefix = "_".join(str(sw) for sw in sub)
        vswitches = []  # store to connect trunks
        for vlan, interfaces in vlans.items():
            # create a virtual switch
            vswitch_id = "vswitch%s" % vswitch_id_counter.next()
            vswitch = g_vlans.add_node(vswitch_id)
            vswitch.device_type = "switch"
            vswitch.device_subtype = "virtual"
            vswitches.append(vswitch)
            # TODO: layout based on midpoint of previous?
            # or if same number as real switches, use their co-ordinates?
            # and then check for coincident?
            vswitch.x = sum(
                i.node['phy'].x for i in interfaces) / len(interfaces) + 50
            vswitch.y = sum(
                i.node['phy'].y for i in interfaces) / len(interfaces) + 50
            vswitch.vlan = vlan

            g_vlan_trunk.add_node(vswitch)

            # and connect from vswitch to the interfaces
            edges_to_add = [(vswitch, iface) for iface in interfaces]
            g_l2.add_edges_from(edges_to_add)

        # remove the physical switches
        g_l2.remove_nodes_from(sub)

        # Note: we don't store the interface names as ciuld clobber
        # eg came from two physical switches, each on gige0
        # if need, work backwards from the router iface and its connectivity

        # and add the trunks
        # TODO: these need annotations!
        # create trunks
        edges_to_add = list(itertools.combinations(vswitches, 2))
        # TODO: ensure only once
        # TODO: filter so only one direction
        g_l2.add_edges_from(edges_to_add, trunk=True)
        g_vlan_trunk.add_edges_from(edges_to_add, trunk=True)


from autonetkit.nidb.device_model import DeviceModel as DeviceModel
from autonetkit.nidb.node import DmNode as DmNode
from autonetkit.nidb.edge import DmEdge as DmEdge
from autonetkit.nidb.interface import DmInterface as DmInterface
from autonetkit.nidb.config_stanza import ConfigStanza as ConfigStanza


import autonetkit.ank as ank

def network_hostname(node):
    network = node.Network
    if network:
        return "%s_%s" % (ank.name_folder_safe(network),
                ank.name_folder_safe(node.label))
    else:
        return ank.name_folder_safe(node.label)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import itertools
from collections import namedtuple
import autonetkit

import autonetkit.log as log
import networkx as nx
from ank_utils import unwrap_graph, unwrap_nodes
from autonetkit.anm import NmEdge, NmNode

# helper namedtuples - until have a more complete schema (such as from Yang)
static_route_v4 = namedtuple("static_route_v4",
                             ["prefix", "netmask", "nexthop", "metric"])

static_route_v6 = namedtuple("static_route_v6",
                             ["prefix", "nexthop", "metric"])

# TODO: add ability to specify labels to unwrap too

# TODO: split into a utils module


def sn_preflen_to_network(address, prefixlen):
    """Workaround for creating an IPNetwork from an address and a prefixlen
    TODO: check if this is part of netaddr module
    """

    import netaddr
    return netaddr.IPNetwork('%s/%s' % (address, prefixlen))


def fqdn(node):
    """

    Example:

    >>> anm = autonetkit.topos.house()
    >>> r1 = anm['phy'].node("r1")
    >>> fqdn(r1)
    'r1.1'

    """
    return '%s.%s' % (node.label, node.asn)


def name_folder_safe(foldername):
    for illegal_char in [' ', '/', '_', ',', '.', '&amp;', '-', '(', ')', ]:
        foldername = foldername.replace(illegal_char, '_')

    # Don't want double _

    while '__' in foldername:
        foldername = foldername.replace('__', '_')
    return foldername


def set_node_default(nm_graph, nbunch=None, **kwargs):
    """Sets all nodes in nbunch to value if key not already set
    Note: this won't apply to future nodes added

    >>> anm = autonetkit.topos.house()
    >>> g_phy = anm['phy']
    >>> r1 = g_phy.node("r1")
    >>> r1.color = "blue"
    >>> [(n, n.color) for n in g_phy]
    [(r4, None), (r5, None), (r1, 'blue'), (r2, None), (r3, None)]
    >>> set_node_default(g_phy, color="red")
    >>> [(n, n.color) for n in g_phy]
    [(r4, 'red'), (r5, 'red'), (r1, 'blue'), (r2, 'red'), (r3, 'red')]

    Can also set for a specific bunch of nodes

    >>> nodes = ["r1", "r2", "r3"]
    >>> set_node_default(g_phy, nodes, role="core")
    >>> [(n, n.role) for n in g_phy]
    [(r4, None), (r5, None), (r1, 'core'), (r2, 'core'), (r3, 'core')]

    """

    # work with the underlying NetworkX graph for efficiency
    graph = unwrap_graph(nm_graph)
    if nbunch is None:
        nbunch = graph.nodes()
    else:
        nbunch = unwrap_nodes(nbunch)
    for node in nbunch:
        for (key, val) in kwargs.items():
            if key not in graph.node[node]:
                graph.node[node][key] = val

# TODO: also add ability to copy multiple attributes

# TODO: rename to copy_node_attr_from


def copy_attr_from(overlay_src, overlay_dst, src_attr, dst_attr=None,
                   nbunch=None, type=None, default=None):
    """

    >>> anm = autonetkit.topos.house()
    >>> g_in = anm['input']
    >>> g_phy = anm['phy']
    >>> [n.color for n in g_phy]
    [None, None, None, None, None]
    >>> set_node_default(g_in, color="red")
    >>> copy_attr_from(g_in, g_phy, "color")
    >>> [n.color for n in g_phy]
    ['red', 'red', 'red', 'red', 'red']

    Can specify a default value if unset

    >>> nodes = ["r1", "r2", "r3"]
    >>> set_node_default(g_in, nodes, role="core")
    >>> copy_attr_from(g_in, g_phy, "role", default="edge")
    >>> [(n, n.role) for n in g_phy]
    [(r4, 'edge'), (r5, 'edge'), (r1, 'core'), (r2, 'core'), (r3, 'core')]


    Can specify the remote attribute to set

    >>> copy_attr_from(g_in, g_phy, "role", "device_role", default="edge")

    Can specify the type to cast to

    >>> g_in.update(memory = "32")
    >>> copy_attr_from(g_in, g_phy, "memory", type=int)
    >>> [n.memory for n in g_phy]
    [32, 32, 32, 32, 32]


    Supported types to case to are float and int

    """

    if not dst_attr:
        dst_attr = src_attr

    graph_src = unwrap_graph(overlay_src)
    graph_dst = unwrap_graph(overlay_dst)
    if not nbunch:
        nbunch = graph_src.nodes()

    for node in nbunch:
        try:
            val = graph_src.node[node].get(src_attr, default)
        except KeyError:

            # TODO: check if because node doesn't exist in dest, or because
            # attribute doesn't exist in graph_src

            log.debug('Unable to copy node attribute %s for %s in %s',
                      src_attr, node, overlay_src)
        else:

            # TODO: use a dtype to take an int, float, etc

            if type is float:
                val = float(val)
            elif type is int:
                val = int(val)

            if node in graph_dst:
                graph_dst.node[node][dst_attr] = val


def copy_int_attr_from(overlay_src, overlay_dst, src_attr, dst_attr=None,
                       nbunch=None, type=None, default=None):

    #TODO: check if copies to loopbacks as well
    """

    >>> anm = autonetkit.topos.house()
    >>> g_in = anm['input']
    >>> g_phy = anm['phy']
    >>> [iface.ospf_cost for node in g_phy for iface in node]
    [None, None, None, None, None, None, None, None, None, None, None, None]
    >>> for node in g_in:
    ...      for interface in node:
    ...         interface.ospf_cost = 10
    >>> copy_int_attr_from(g_in, g_phy, "ospf_cost")
    >>> [iface.ospf_cost for node in g_phy for iface in node]
    [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]


    Supported types to case to are float and int

    """

    # note; uses high-level API for practicality over raw speed

    if not dst_attr:
        dst_attr = src_attr

    if not nbunch:
        nbunch = overlay_src.nodes()

    for node in nbunch:
        for src_int in node:
            val = src_int.get(src_attr)
            if val is None:
                val = default

            if type is float:
                val = float(val)
            elif type is int:
                val = int(val)

            if node not in overlay_dst:
                continue

            dst_int = overlay_dst.interface(src_int)
            if dst_int is not None:
                dst_int.set(dst_attr, val)


def copy_edge_attr_from(overlay_src, overlay_dst, src_attr,
                        dst_attr=None, type=None, default=None):
    # note this won't work if merge/aggregate edges

    if not dst_attr:
        dst_attr = src_attr

    for edge in overlay_src.edges():
        try:
            val = edge.get(src_attr)
            if val is None:
                val = default
        except KeyError:

            # TODO: check if because edge doesn't exist in dest, or because
            # attribute doesn't exist in graph_src

            log.debug('Unable to copy edge attribute %s for (%s, %s) in %s',
                      src_attr, edge.src, edge.dst, overlay_src)

        else:

            # TODO: use a dtype to take an int, float, etc

            if type is float:
                val = float(val)
            elif type is int:
                val = int(val)

            try:
                overlay_dst.edge(edge).set(dst_attr, val)
            except AttributeError:
                # fail to debug - as attribute may not have been set
                log.debug('Unable to set edge attribute on %s in %s',
                          edge, overlay_dst)


# TODO: make edges own module

def wrap_edges(nm_graph, edges):
    """ wraps edge ids into edge overlay

    >>> anm = autonetkit.topos.house()
    >>> g_phy = anm['phy']
    >>> elist = [("r1", "r2"), ("r2", "r3")]
    >>> edges = wrap_edges(g_phy, elist)

    The edges are now NetworkModel edge objects

    >>> edges
    [(r1, r2), (r2, r3)]

    """

    # TODO: make support multigraphs

    edges = list(edges)
    if not any(len(e) for e in edges):
        return []  # each edge tuple is empty

    try:

        # strip out data from (src, dst, data) tuple

        edges = [(s, t) for (s, t, _) in edges]
    except ValueError:
        pass  # already of form (src, dst)

    return list(NmEdge(nm_graph._anm, nm_graph._overlay_id, src, dst)
                for (src, dst) in edges)


def wrap_nodes(nm_graph, nodes):
    """ wraps node id into node overlay

    >>> anm = autonetkit.topos.house()
    >>> g_phy = anm['phy']
    >>> nlist = ["r1", "r2", "r3"]
    >>> nodes = wrap_nodes(g_phy, nlist)

    The nodes are now NetworkModel node objects

    >>> nodes
    [r1, r2, r3]

    This is generally used in internal functions.
    An alternative method is:

    >>> [g_phy.node(n) for n in nlist]
    [r1, r2, r3]

     """

    return [NmNode(nm_graph._anm, nm_graph._overlay_id, node)
            for node in nodes]


def in_edges(nm_graph, nodes=None):

    # TODO: make support multigraphs

    graph = unwrap_graph(nm_graph)
    edges = graph.in_edges(nodes)
    return wrap_edges(nm_graph, edges)


def split(nm_graph, edges, retain=None, id_prepend=''):
    """
    Splits edges in two, retaining any attributes specified.

    >>> anm = autonetkit.topos.house()
    >>> g_phy = anm['phy']
    >>> edge = g_phy.edge("r1", "r2")
    >>> new_nodes = split(g_phy, edge)
    >>> new_nodes
    [r1_r2]
    >>> [n.neighbors() for n in new_nodes]
    [[r1, r2]]


    For multiple edges and specifying a prepend for the new nodes

    >>> anm = autonetkit.topos.house()
    >>> g_phy = anm['phy']
    >>> edges = g_phy.node("r2").edges()
    >>> new_nodes = split(g_phy, edges, id_prepend="split_")
    >>> new_nodes
    [split_r2_r4, split_r1_r2, split_r2_r3]
    >>> [n.neighbors() for n in new_nodes]
    [[r4, r2], [r1, r2], [r2, r3]]

    """

    if retain is None:
        retain = []

    try:
        # TODO: find more efficient operation to test if string-like
        retain.lower()
        retain = [retain]  # was a string, put into list
    except AttributeError:
        pass  # already a list

    graph = unwrap_graph(nm_graph)
    edges_to_add = []
    added_nodes = []

    # handle single edge
    if edges in nm_graph.edges():
        edges = [edges] # place into list for iteration

    edges = list(edges)

    for edge in edges:
        src = edge.src
        dst = edge.dst

        # Form ID for the new node

        if graph.is_directed():
            new_id = '%s%s_%s' % (id_prepend, src, dst)
        else:

            # undirected, make id deterministic across ank runs

            # use sorted for consistency
            (node_a, node_b) = sorted([src, dst])
            new_id = '%s%s_%s' % (id_prepend, node_a, node_b)

        if nm_graph.is_multigraph():
            new_id = new_id + '_%s' % edge.ekey

        ports = edge.raw_interfaces
        data = edge._data
        src_data = data.copy()
        if src in ports:
            src_int_id = ports[src.node_id]
            src_data['_ports'] = {src.node_id: src_int_id}
        dst_data = data.copy()
        if dst in ports:
            dst_int_id = ports[dst.node_id]
            dst_data['_ports'] = {dst.node_id: dst_int_id}

        # Note: don't retain ekey since adding to a new node

        append = (src.node_id, new_id, src_data)
        edges_to_add.append(append)
        append = (dst.node_id, new_id, dst_data)
        edges_to_add.append(append)

        added_nodes.append(new_id)

    nm_graph.add_nodes_from(added_nodes)
    nm_graph.add_edges_from(edges_to_add)

    # remove the pre-split edges

    nm_graph.remove_edges_from(edges)

    return wrap_nodes(nm_graph, added_nodes)


def explode_nodes(nm_graph, nodes, retain=None):
    """Explodes all nodes in nodes
    TODO: explain better
    TODO: Add support for digraph - check if nm_graph.is_directed()

    >>> anm = autonetkit.topos.mixed()
    >>> g_phy = anm['phy']
    >>> switches = g_phy.switches()
    >>> exploded_edges = explode_nodes(g_phy, switches)
    >>> exploded_edges
    [(r1, r2)]


    Or to explode a specific node

    >>> anm = autonetkit.topos.house()
    >>> g_phy = anm['phy']
    >>> g_phy.nodes()
    [r4, r5, r1, r2, r3]

    >>> sorted(g_phy.edges())
    [(r1, r2), (r1, r3), (r2, r3), (r4, r2), (r4, r5), (r5, r3)]

    >>> r2 = g_phy.node("r2")
    >>> exploded_edges = explode_nodes(g_phy, r2)
    >>> exploded_edges
    [(r1, r4), (r3, r4), (r1, r3)]
    >>> g_phy.nodes()
    [r4, r5, r1, r3]

    >>> sorted(g_phy.edges())
    [(r1, r3), (r4, r1), (r4, r3), (r4, r5), (r5, r3)]

    """
    if retain is None:
        retain = []

    log.debug('Exploding nodes')
    try:
        retain.lower()
        retain = [retain]  # was a string, put into list
    except AttributeError:
        pass  # already a list

    total_added_edges = []  # keep track to return

    if nodes in nm_graph:
        nodes = [nodes] # place into list for iteration

    for node in nodes:

        edges = node.edges()
        edge_pairs = [(e1, e2) for e1 in edges for e2 in edges if e1
                      != e2]
        added_pairs = set()
        for edge_pair in edge_pairs:
            (src_edge, dst_edge) = sorted(edge_pair)
            if (src_edge, dst_edge) in added_pairs:
                continue  # already added this link pair in other direction
            else:
                added_pairs.add((src_edge, dst_edge))

            src = src_edge.dst  # src is the exploded node
            dst = dst_edge.dst  # src is the exploded node

            if src == dst:
                continue  # don't add self-loop

            data = dict((key, src_edge._data.get(key)) for key in
                        retain)
            node_to_dst_data = dict((key, dst_edge._data.get(key))
                                    for key in retain)
            data.update(node_to_dst_data)

            data['_ports'] = {}
            try:
                src_int_id = src_edge.raw_interfaces[src.node_id]
            except KeyError:
                pass  # not set
            else:
                data['_ports'][src.node_id] = src_int_id

            try:
                dst_int_id = dst_edge.raw_interfaces[dst.node_id]
            except KeyError:
                pass  # not set
            else:
                data['_ports'][dst.node_id] = dst_int_id

            new_edge = (src.node_id, dst.node_id, data)

            # TODO: use add_edge

            nm_graph.add_edges_from([new_edge])
            total_added_edges.append(new_edge)

        nm_graph.remove_node(node)
    return wrap_edges(nm_graph, total_added_edges)


def label(nm_graph, nodes):
    return list(nm_graph._anm.node_label(node) for node in nodes)


def connected_subgraphs(nm_graph, nodes = None):
    """

    >>> anm = autonetkit.topos.house()
    >>> g_phy = anm['phy']
    >>> connected_subgraphs(g_phy)
    [[r4, r5, r1, r2, r3]]
    >>> edges = [("r2", "r4"), ("r3", "r5")]
    >>> g_phy.remove_edges_from(edges)
    >>> connected_subgraphs(g_phy)
    [[r1, r2, r3], [r4, r5]]


    """
    if nodes is None:
        nodes = nm_graph.nodes()
    else:
        nodes = list(unwrap_nodes(nodes))
    graph = unwrap_graph(nm_graph)
    subgraph = graph.subgraph(nodes)
    if not len(subgraph.edges()):

        # print "Nothing to aggregate for %s: no edges in subgraph"

        pass
    if graph.is_directed():
        component_nodes_list = \
            nx.strongly_connected_components(subgraph)
    else:
        component_nodes_list = nx.connected_components(subgraph)

    wrapped = []
    for component in component_nodes_list:
        wrapped.append(list(wrap_nodes(nm_graph, component)))

    return wrapped


def aggregate_nodes(nm_graph, nodes, retain=None):
    """Combines connected into a single node"""
    if retain is None:
        retain = []

    try:
        retain.lower()
        retain = [retain]  # was a string, put into list
    except AttributeError:
        pass  # already a list

    nodes = list(unwrap_nodes(nodes))
    graph = unwrap_graph(nm_graph)
    subgraph = graph.subgraph(nodes)
    if not len(subgraph.edges()):

        # print "Nothing to aggregate for %s: no edges in subgraph"

        pass
    total_added_edges = []
    if graph.is_directed():
        component_nodes_list = \
            nx.strongly_connected_components(subgraph)
    else:
        component_nodes_list = nx.connected_components(subgraph)
    for component_nodes in component_nodes_list:
        if len(component_nodes) > 1:
            component_nodes = [nm_graph.node(n)
                               for n in component_nodes]

            # TODO: could choose most connected, or most central?
            # TODO: refactor so use nodes_to_remove

            nodes_to_remove = list(component_nodes)
            base = nodes_to_remove.pop()  # choose a base device to retain
            log.debug('Retaining %s, removing %s', base,
                      nodes_to_remove)

            external_edges = []
            for node in nodes_to_remove:
                external_edges += [e for e in node.edges() if e.dst
                                   not in component_nodes]
                # all edges out of component

            log.debug('External edges %s', external_edges)
            edges_to_add = []
            for edge in external_edges:
                dst = edge.dst
                data = dict((key, edge._data.get(key)) for key in
                            retain)
                ports = edge.raw_interfaces
                dst_int_id = ports[dst.node_id]

                # TODO: bind to (and maybe add) port on the new switch?

                data['_ports'] = {dst.node_id: dst_int_id}

                append = (base.node_id, dst.node_id, data)
                edges_to_add.append(append)

            nm_graph.add_edges_from(edges_to_add)
            total_added_edges += edges_to_add
            nm_graph.remove_nodes_from(nodes_to_remove)

    return wrap_edges(nm_graph, total_added_edges)


def most_frequent(iterable):
    """returns most frequent item in iterable"""

# from http://stackoverflow.com/q/1518522

    gby = itertools.groupby
    try:
        return max(gby(sorted(iterable)), key=lambda (x, v):
                   (len(list(v)), -iterable.index(x)))[0]
    except ValueError, error:
        log.warning('Unable to calculate most_frequent, %s', error)
        return None


def neigh_most_frequent(nm_graph, node, attribute,
                        attribute_graph=None, allow_none=False):
    """Used to explicitly force most frequent -
    useful if integers such as ASN which would otherwise return mean"""

    # TODO: rename to median?

    graph = unwrap_graph(nm_graph)
    if attribute_graph:
        attribute_graph = unwrap_graph(attribute_graph)
    else:
        attribute_graph = graph  # use input graph
    node = unwrap_nodes(node)
    values = [attribute_graph.node[n].get(attribute) for n in
              graph.neighbors(node)]
    values = sorted(values)
    if not allow_none:
        values = [v for v in values if v is not None]
    return most_frequent(values)


def neigh_average(nm_graph, node, attribute, attribute_graph=None):
    """
    averages out attribute from neighbors in specified nm_graph
    attribute_graph is the graph to read the attribute from
    if property is numeric, then return mean
    else return most frequently occuring value
    """

    graph = unwrap_graph(nm_graph)
    if attribute_graph:
        attribute_graph = unwrap_graph(attribute_graph)
    else:
        attribute_graph = graph  # use input graph
    node = unwrap_nodes(node)
    values = [attribute_graph.node[n].get(attribute) for n in
              graph.neighbors(node)]

# TODO: use neigh_attr

    try:
        values = [float(val) for val in values]
        return sum(values) / len(values)
    except ValueError:
        return most_frequent(values)


def neigh_attr(nm_graph, node, attribute, attribute_graph=None):
    """TODO:
    tidy up parameters to take attribute_graph first, and
    then evaluate if attribute_graph set, if not then use attribute_graph
    as attribute
    explain how nm_graph and attribute_graph work, eg for G_ip and
    G_phy
    """

    graph = unwrap_graph(nm_graph)
    node = unwrap_nodes(node)
    if attribute_graph:
        attribute_graph = unwrap_graph(attribute_graph)
    else:
        attribute_graph = graph  # use input graph

    # Only look at nodes which exist in attribute_graph

    neighs = (n for n in graph.neighbors(node))
    valid_nodes = (n for n in neighs if n in attribute_graph)
    return (attribute_graph.node[node].get(attribute) for node in
            valid_nodes)


def neigh_equal(nm_graph, node, attribute, attribute_graph=None):
    """Boolean, True if neighbors in nm_graph
    all have same attribute in attribute_graph"""

    neigh_attrs = neigh_attr(nm_graph, node, attribute, attribute_graph)
    return len(set(neigh_attrs)) == 1


def unique_attr(nm_graph, attribute):
    graph = unwrap_graph(nm_graph)
    return set(graph.node[node].get(attribute) for node in graph)


def groupby(attribute, nodes):
    """Takes a group of nodes and returns a generator of (attribute, nodes)
     for each attribute value A simple wrapped around itertools.groupby
     that creates a lambda for the attribute
    """

    keyfunc = lambda x: x.get(attribute)
    nodes = sorted(nodes, key=keyfunc)
    return itertools.groupby(nodes, key=keyfunc)

def shortest_path(nm_graph, src, dst):

    # TODO: move to utils
# TODO: use networkx boundary nodes directly: does the same thing

    graph = unwrap_graph(nm_graph)
    src_id = unwrap_nodes(src)
    dst_id = unwrap_nodes(dst)

    #TODO: check path works for muli-edge graphs too
    path = nx.shortest_path(graph, src_id, dst_id)

    return wrap_nodes(nm_graph, path)


def boundary_nodes(nm_graph, nodes):
    """ returns nodes at boundary of G based on
    edge_boundary from networkx """

    # TODO: move to utils
# TODO: use networkx boundary nodes directly: does the same thing

    graph = unwrap_graph(nm_graph)
    nodes = list(nodes)
    nbunch = list(unwrap_nodes(nodes))

    # find boundary

    b_edges = nx.edge_boundary(graph, nbunch)  # boundary edges
    internal_nodes = [s for (s, _) in b_edges]
    assert all(n in nbunch for n in internal_nodes)  # check internal

    return wrap_nodes(nm_graph, internal_nodes)



import logging
from autonetkit.nidb.node import DmNode
from autonetkit.nidb.interface import DmInterface
from autonetkit.log import CustomAdapter
import autonetkit.log as log


class DmEdge(object):

    """API to access edge in nidb"""

    def __init__(self, nidb, src_id, dst_id, ekey=0):
        # Set using this method to bypass __setattr__
        object.__setattr__(self, 'nidb', nidb)
        object.__setattr__(self, 'src_id', src_id)
        object.__setattr__(self, 'dst_id', dst_id)
        object.__setattr__(self, 'ekey', ekey)  # for multigraphs
        #logger = logging.getLogger("ANK")
        #logstring = "Edge: %s" % str(self)
        #self.log = CustomAdapter(logger, {'item': logstring})
        logger = log
        object.__setattr__(self, 'log', logger)

    def __repr__(self):
        if self.is_multigraph():
            return '(%s, %s, %s)' % (self.src,
                                     self.dst, self.ekey)

        return '(%s, %s)' % (self.src, self.dst)

    @property
    def raw_interfaces(self):
        """Direct access to the interfaces dictionary, used by ANK modules"""
        return self._ports

    @raw_interfaces.setter
    def raw_interfaces(self, value):
        self._ports = value

    def is_multigraph(self):
        return self._graph.is_multigraph()

    @property
    def src(self):
        return DmNode(self.nidb, self.src_id)

    @property
    def src_int(self):
        src_int_id = self._ports[self.src_id]
        return DmInterface(self.nidb, self.src_id, src_int_id)

    @property
    def dst_int(self):
        dst_int_id = self._ports[self.dst_id]
        return DmInterface(self.nidb, self.dst_id, dst_int_id)

    def __eq__(self, other):
        """"""
        if self.is_multigraph():
            try:
                if other.is_multigraph():
                    return (self.src_id, self.dst_id, self.ekey) == (other.src_id,
                                                                     other.dst_id, other.ekey)
                else:
                    # multi, single
                    return (self.src_id, self.dst_id) == (other.src_id,
                                                          other.dst_id)

            except AttributeError:
                if len(other) == 2:
                    # (src, dst)
                    return (self.src_id, self.dst_id) == other
                elif len(other) == 3:
                    # (src, dst, key)
                    return (self.src_id, self.dst_id, self.ekey) == other

        try:
            # self is single, other is single or multi -> only compare (src,
            # dst)
            return (self.src_id, self.dst_id) == (other.src_id, other.dst_id)
        except AttributeError:
            # compare to strings
            return (self.src_id, self.dst_id) == other

    def __lt__(self, other):
        """"""
        if self.is_multigraph() and other.is_multigraph():
            return (self.src.node_id, self.dst.node_id, self.ekey) \
                < (other.src.node_id, other.dst.node_id, other.ekey)

        return (self.src.node_id, self.dst.node_id) \
            < (other.src.node_id, other.dst.node_id)

    # Internal properties
    def __nonzero__(self):
        """Allows for checking if edge exists
        """
        if self.is_multigraph():
            return self._graph.has_edge(self.src_id, self.dst_id,
                                        key=self.ekey)

        return self._graph.has_edge(self.src_id, self.dst_id)

    @property
    def dst(self):
        return DmNode(self.nidb, self.dst_id)

    def dump(self):
        return str(self._graph[self.src_id][self.dst_id])

    def __nonzero__(self):
        """Allows for checking if edge exists
        """
        try:
            # TODO: refactor to be _graph.has_edge(src, dst)
            _ = self._graph[self.src_id][self.dst_id]
            return True
        except KeyError:
            return False

    @property
    def _data(self):
        """Return data the node belongs to"""
        if self.is_multigraph():
            return self._graph[self.src_id][self.dst_id][self.ekey]

        return self._graph[self.src_id][self.dst_id]

    @property
    def _graph(self):
        """Return graph the node belongs to"""
        return self.nidb.raw_graph()

    def get(self, key):
        """For consistency, edge.get(key) is neater than getattr(edge, key)"""
        return self.__getattr__(key)

    def __getattr__(self, key):
        """Returns edge property"""
        return self._data.get(key)

    def __setattr__(self, key, val):
        """Sets edge property"""
        self._data[key] = val



#!/usr/bin/python
# -*- coding: utf-8 -*-
import re
import time

import autonetkit.ank_messaging as ank_messaging
import autonetkit.config as config
import autonetkit.log as log

try:
    import Exscript
except ImportError:
    log.warning('Deployment requires Exscript: pip install https://github.com/knipknap/exscript/tarball/master'
                )


def deploy(host, username, dst_folder, key_filename=None,
    parallel_count=5):
    tar_file = package(dst_folder)
    transfer(host, username, tar_file, key_filename=key_filename)
    extract(
        host,
        username,
        tar_file,
        dst_folder,
        key_filename=key_filename,
        parallel_count=parallel_count,
        )


def package(src_dir, target='netkit_lab'):
    log.info('Packaging %s' % src_dir)
    import tarfile
    import os
    tar_filename = '%s.tar.gz' % target
    tar = tarfile.open(os.path.join(tar_filename), 'w:gz')
    tar.add(src_dir)
    tar.close()
    return tar_filename


def transfer(host, username, local, remote=None, key_filename=None):
    log.debug('Transferring lab to %s' % host)
    log.info('Transferring Netkit lab')
    if not remote:
        remote = local  # same filename
    import paramiko

    # import logging
    # logging.getLogger("paramiko").setLevel(logging.DEBUG)

    ssh = paramiko.SSHClient()

    # ssh.set_log_channel("ANK")

    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    if key_filename:
        log.debug('Connecting to %s with %s and key %s' % (host,
                  username, key_filename))
        ssh.connect(host, username=username, key_filename=key_filename)
    else:
        log.info('Connecting to %s with %s' % (host, username))
        ssh.connect(host, username=username)
    log.info('Opening SSH for SFTP')
    ftp = ssh.open_sftp()
    log.info('Putting file %s tspoto %s' % (local, remote))
    ftp.put(local, remote)
    log.info('Put file %s to %s' % (local, remote))
    ftp.close()


def extract(
    host,
    username,
    tar_file,
    cd_dir,
    timeout=45,
    key_filename=None,
    verbosity=0,
    parallel_count=5,
    ):
    """Extract and start lab"""

    log.debug('Extracting and starting lab on %s' % host)
    log.info('Extracting and starting Netkit lab')
    from Exscript import Account
    from Exscript.util.start import start
    from Exscript.util.match import first_match
    from Exscript import PrivateKey
    from Exscript.protocols.Exception import InvalidCommandException

    messaging = ank_messaging

    def starting_host(protocol, index, data):
        log.info('Starting %s' % data.group(1))

    def lab_started(protocol, index, data):
        log.info('Lab started on %s' % host)

    def make_not_found(protocol, index, data):
        log.warning('Make not installed on remote host %s. Please install make and retry.'
                     % host)
        return

    def process_vlist(response):
        """Obtain VM to PID listing: required if terminating a numeric VM"""

        # TODO: could process using textfsm template

        vm_to_pid = {}
        for line in response.splitlines():
            match = re.match(r'^\w+\s+(\w+)\s+(\d+)', line)
            if match:
                vm = match.group(1)
                pid = match.group(2)
                vm_to_pid[vm] = pid

        return vm_to_pid

    def start_lab(thread, host, conn):
        conn.set_timeout(timeout)
        conn.add_monitor(r'Starting "(\w+)"', starting_host)
        conn.add_monitor(r'The lab has been started', lab_started)
        lab_vlist = []

        # conn.add_monitor(r'Virtual machine "((\S*_*)+)" is already running. Please', already_running_b)

        conn.add_monitor(r'make: not found', make_not_found)

        # conn.data_received_event.connect(data_received)

        conn.execute('cd %s' % cd_dir)
        conn.execute('lcrash -k')
        conn.execute('lclean')
        conn.execute('cd')  # back to home directory tar file copied to
        conn.execute('tar -xzf %s' % tar_file)
        conn.execute('cd %s' % cd_dir)

        conn.execute('linfo')
        linfo_response = str(conn.response)
        vm_list = []
        for line in linfo_response.splitlines():
            if 'The lab is made up of' in line:
                open_bracket = line.index('(')
                close_bracket = line.index(')')
                vm_list = line[open_bracket + 1:close_bracket]
                vm_list = vm_list.split()
                log.info('The lab contains VMs %s' % ', '.join(vm_list))

        # now check if any vms are still running

        conn.execute('vlist')
        response = str(conn.response)
        lab_vlist = process_vlist(response)

        for virtual_machine in lab_vlist:
            if virtual_machine in vm_list:
                if virtual_machine.isdigit:

                    # convert to PID if numeric, as vcrash can't crash numeric ids (treats as PID)

                    crash_id = lab_vlist.get(virtual_machine)
                else:
                    crash_id = virtual_machine  # use name

                if crash_id:

                    # crash_id may not be set, if machine not present in initial vlist, if so then ignore

                    log.info('Stopping running VM %s' % virtual_machine)
                    conn.execute('vcrash %s' % crash_id)

        conn.execute('vlist')
        conn.execute('lclean')
        start_command = 'lstart -p%s -o --con0=none' % parallel_count
        lab_is_started = False
        while lab_is_started == False:
            try:
                log.info('Starting lab')
                conn.execute(start_command)
            except InvalidCommandException, error:
                error_string = str(error)
                if 'already running' in error_string:

                    conn.execute('vlist')
                    response = str(conn.response)
                    lab_vlist = process_vlist(response)

                    running_vms = []
                    for line in error_string.splitlines():
                        if 'already running' in line:
                            running_vm = line.split('"')[1]
                            running_vms.append(running_vm)

                    for virtual_machine in running_vms:
                        if virtual_machine.isdigit:

                            # convert to PID if numeric, as vcrash can't crash numeric ids (treats as PID)

                            crash_id = lab_vlist.get(virtual_machine)
                        else:
                            crash_id = virtual_machine  # use name

                        if crash_id:

                            # crash_id may not be set, if machine not present in initial vlist, if so then ignore

                            log.info('Stopping running VM %s'
                                    % virtual_machine)
                            conn.execute('vcrash %s' % crash_id)

                    time.sleep(1)
            else:

                    # conn.execute(start_command)

                lab_is_started = True
        first_match(conn, r'^The lab has been started')
        log.info('Lab started')  # TODO: make this captured - need to debug capturing
        conn.send('exit')

    if key_filename:
        key = PrivateKey.from_file(key_filename)
        log.debug('Connecting to %s with username %s and key %s'
                  % (host, username, key_filename))
        accounts = [Account(username, key=key)]
    else:
        log.debug('Connecting to %s with username %s' % (host,
                  username))
        accounts = [Account(username)]

    hosts = ['ssh://%s' % host]
    verbosity = -1
    start(accounts, hosts, start_lab, verbose=verbosity)



#!/usr/bin/python
# -*- coding: utf-8 -*-
import autonetkit.ank as ank_utils
import autonetkit.config
import autonetkit.log as log

SETTINGS = autonetkit.config.settings


# TODO: refactor to go in chronological workflow order




#@call_log
def build_ip(anm):
    g_ip = anm.add_overlay('ip')
    g_l2_bc = anm['layer2_bc']
    g_phy = anm['phy']
    # Retain arbitrary ASN allocation for IP addressing
    g_ip.add_nodes_from(g_l2_bc, retain=["asn", "broadcast_domain"])
    g_ip.add_edges_from(g_l2_bc.edges())

    #TODO:
    for bc in g_ip.nodes("broadcast_domain"):
        bc.allocate = True

    for bc in g_ip.nodes("broadcast_domain"):
        # Encapsulated if any neighbor interface has
        for edge in bc.edges():
            if edge.dst_int['phy'].l2_encapsulated:
                log.debug("Removing IP allocation for broadcast_domain %s "
                         "as neighbor %s is L2 encapsulated", bc, edge.dst)

                #g_ip.remove_node(bc)
                bc.allocate = False

                # and mark on connected interfaces
                for neigh_int in bc.neighbor_interfaces():
                    neigh_int.allocate = False

                break


    # copy over skipped loopbacks
    #TODO: check if loopbck copy attr
    for node in g_ip.l3devices():
        for interface in node.loopback_interfaces():
            if interface['phy'].allocate is not None:
                interface['ip'].allocate = interface['phy'].allocate


def build_ipv4(anm, infrastructure=True):
    import autonetkit.design.ip_addressing.ipv4
    autonetkit.design.ip_addressing.ipv4.build_ipv4(
        anm, infrastructure=infrastructure)

def build_ipv6(anm):
    #TODO: check why ipv6 doesn't take infrastructure
    import autonetkit.design.ip_addressing.ipv6
    autonetkit.design.ip_addressing.ipv6.build_ipv6(anm)
#@call_log



import autonetkit.ank as ank_utils
import autonetkit.log as log
from autonetkit.ank_utils import call_log

#@call_log


def build_ibgp_v4(anm):
    # TODO: remove the bgp layer and have just ibgp and ebgp
    # TODO: build from design rules, currently just builds from ibgp links in bgp layer
    # TODO: base on generic ibgp graph, rather than bgp graph
    g_bgp = anm['bgp']
    g_phy = anm['phy']
    g_ibgpv4 = anm.add_overlay("ibgp_v4", directed=True)
    ipv4_nodes = set(g_phy.routers("use_ipv4"))
    if len(ipv4_nodes) == 0:
        return
    g_ibgpv4.add_nodes_from((n for n in g_bgp if n in ipv4_nodes),
                            retain=["ibgp_role", "hrr_cluster", "rr_cluster"])
    g_ibgpv4.add_edges_from(g_bgp.edges(type="ibgp"), retain="direction")

#@call_log


def build_ibgp_v6(anm):
    # TODO: remove the bgp layer and have just ibgp and ebgp
    # TODO: build from design rules, currently just builds from ibgp links in bgp layer
    # TODO: base on generic ibgp graph, rather than bgp graph
    g_bgp = anm['bgp']
    g_phy = anm['phy']
    g_ibgpv6 = anm.add_overlay("ibgp_v6", directed=True)
    ipv6_nodes = set(g_phy.routers("use_ipv6"))
    if len(ipv6_nodes) == 0:
        return
    g_ibgpv6.add_nodes_from((n for n in g_bgp if n in ipv6_nodes),
                            retain=["ibgp_role", "hrr_cluster", "rr_cluster"])
    g_ibgpv6.add_edges_from(g_bgp.edges(type="ibgp"), retain="direction")

#@call_log


def build_ebgp_v4(anm):
    # TODO: remove the bgp layer and have just ibgp and ebgp
    # TODO: build from design rules, currently just builds from ibgp links in
    # bgp layer
    g_ebgp = anm['ebgp']
    g_phy = anm['phy']
    g_ebgpv4 = anm.add_overlay("ebgp_v4", directed=True)
    ipv4_nodes = set(g_phy.routers("use_ipv4"))
    if len(ipv4_nodes) == 0:
        return
    g_ebgpv4.add_nodes_from(n for n in g_ebgp if n in ipv4_nodes)
    g_ebgpv4.add_edges_from(g_ebgp.edges(), retain="direction")

#@call_log


def build_ebgp_v6(anm):
    # TODO: remove the bgp layer and have just ibgp and ebgp
    # TODO: build from design rules, currently just builds from ibgp links in
    # bgp layer
    g_ebgp = anm['ebgp']
    g_phy = anm['phy']
    g_ebgpv6 = anm.add_overlay("ebgp_v6", directed=True)
    ipv6_nodes = set(g_phy.routers("use_ipv6"))
    if len(ipv6_nodes) == 0:
        return
    g_ebgpv6.add_nodes_from(n for n in g_ebgp if n in ipv6_nodes)
    g_ebgpv6.add_edges_from(g_ebgp.edges(), retain="direction")

#@call_log


def build_ebgp(anm):
    g_l3 = anm['layer3']

    g_ebgp = anm.add_overlay("ebgp", directed=True)

    g_ebgp.add_nodes_from(g_l3.routers())
    ank_utils.copy_int_attr_from(g_l3, g_ebgp, "multipoint")

    ebgp_edges = [e for e in g_l3.edges() if e.src.asn != e.dst.asn]
    g_ebgp.add_edges_from(ebgp_edges, bidirectional=True, type='ebgp')

#@call_log


def build_ibgp(anm):
    g_in = anm['input']
    g_bgp = anm['bgp']

    # TODO: build direct to ibgp graph - can construct combined bgp for vis
    #TODO: normalise input property

    ank_utils.copy_attr_from(g_in, g_bgp, "ibgp_role")
    ank_utils.copy_attr_from(
        g_in, g_bgp, "ibgp_l2_cluster", "hrr_cluster", default=None)
    ank_utils.copy_attr_from(
        g_in, g_bgp, "ibgp_l3_cluster", "rr_cluster", default=None)

    # TODO: add more detailed logging

    for n in g_bgp:
        # Tag with label to make logic clearer
        if n.ibgp_role is None:
            n.ibgp_role = "Peer"

            # TODO: if top-level, then don't mark as RRC

    ibgp_nodes = [n for n in g_bgp if not n.ibgp_role is "Disabled"]

    # Notify user of non-ibgp nodes
    non_ibgp_nodes = [n for n in g_bgp if n.ibgp_role is "Disabled"]
    if 0 < len(non_ibgp_nodes) < 10:
        log.info("Skipping iBGP for iBGP disabled nodes: %s", non_ibgp_nodes)
    elif len(non_ibgp_nodes) >= 10:
        log.info("Skipping iBGP for more than 10 iBGP disabled nodes:"
                 "refer to visualization for resulting topology.")

    # warn for any nodes that have RR set but no rr_cluster, or HRR set and no
    # hrr_cluster
    rr_mismatch = [
        n for n in ibgp_nodes if n.ibgp_role == "RR" and n.rr_cluster is None]
    if len(rr_mismatch):
        log.warning("Some routers are set as RR but have no rr_cluster: %s. Please specify an rr_cluster for peering."
                    % ", ".join(str(n) for n in rr_mismatch))

    hrr_mismatch = [
        n for n in ibgp_nodes if n.ibgp_role == "HRR" and n.hrr_cluster is None]
    if len(hrr_mismatch):
        log.warning("Some routers are set as HRR but have no hrr_cluster: %s. Please specify an hrr_cluster for peering."
                    % ", ".join(str(n) for n in hrr_mismatch))

    for _, asn_devices in ank_utils.groupby("asn", ibgp_nodes):
        asn_devices = list(asn_devices)

        # iBGP peer peers with
        peers = [n for n in asn_devices if n.ibgp_role == "Peer"]
        rrs = [n for n in asn_devices if n.ibgp_role == "RR"]
        hrrs = [n for n in asn_devices if n.ibgp_role == "HRR"]
        rrcs = [n for n in asn_devices if n.ibgp_role == "RRC"]

        over_links = []
        up_links = []
        down_links = []

        # 0. RRCs can only belong to either an rr_cluster or a hrr_cluster
        invalid_rrcs = [r for r in rrcs if r.rr_cluster is not None
                        and r.hrr_cluster is not None]
        if len(invalid_rrcs):
            message = ", ".join(str(r) for r in invalid_rrcs)
            log.warning("RRCs can only have either a rr_cluster or hrr_cluster set. "
                        "The following have both set, and only the rr_cluster will be used: %s", message)

        # TODO: do we also want to warn for RRCs with no cluster set? Do we also exclude these?
        # TODO: do we also want to warn for HRRs and RRs with no cluster set?
        # Do we also exclude these?

        # 1. Peers:
        # 1a. Peers connect over to peers
        over_links += [(s, t) for s in peers for t in peers]
        # 1b. Peers connect over to RRs
        over_links += [(s, t) for s in peers for t in rrs]

        # 2. RRs:
        # 2a. RRs connect over to Peers
        over_links += [(s, t) for s in rrs for t in peers]
        # 2b. RRs connect over to RRs
        over_links += [(s, t) for s in rrs for t in rrs]
        # 2c. RRs connect down to RRCs in same rr_cluster
        down_links += [(s, t) for s in rrs for t in rrcs
                       if s.rr_cluster == t.rr_cluster != None]
        # 2d. RRs connect down to HRRs in the same rr_cluster
        down_links += [(s, t) for s in rrs for t in hrrs
                       if s.rr_cluster == t.rr_cluster != None]

        # 3. HRRs
        # 3a. HRRs connect up to RRs in the same rr_cluster
        up_links += [(s, t) for s in hrrs for t in rrs
                     if s.rr_cluster == t.rr_cluster != None]
        # 3b. HRRs connect down to RRCs in same hrr_cluster (providing RRC has
        # no rr_cluster set)
        down_links += [(s, t) for s in hrrs for t in rrcs
                       if s.hrr_cluster == t.hrr_cluster != None
                       and t.rr_cluster is None]

        # 4. RRCs
        # 4a. RRCs connect up to RRs in the same rr_cluster (regardless if RRC
        # has hrr_cluster set)
        up_links += [(s, t) for s in rrcs for t in rrs
                     if s.rr_cluster == t.rr_cluster != None]
        # 3b. RRCs connect up to HRRs in same hrr_cluster (providing RRC has no
        # rr_cluster set)
        up_links += [(s, t) for s in rrcs for t in hrrs
                     if s.hrr_cluster == t.hrr_cluster != None
                     and s.rr_cluster is None]

        # Remove self-links
        over_links = [(s, t) for s, t in over_links if s != t]
        up_links = [(s, t) for s, t in up_links if s != t]
        down_links = [(s, t) for s, t in down_links if s != t]

        g_bgp.add_edges_from(over_links, type='ibgp', direction='over')
        g_bgp.add_edges_from(up_links, type='ibgp', direction='up')
        g_bgp.add_edges_from(down_links, type='ibgp', direction='down')

#@call_log


def build_bgp(anm):
    """Build iBGP end eBGP overlays"""
    # eBGP
    g_in = anm['input']
    g_l3 = anm['layer3']

    if not anm['phy'].data.enable_routing:
        log.info("Routing disabled, not configuring BGP")
        return

    build_ebgp(anm)
    build_ebgp_v4(anm)
    build_ebgp_v6(anm)

    """TODO: remove from here once compiler updated"""
    g_bgp = anm.add_overlay("bgp", directed=True)
    g_bgp.add_nodes_from(g_l3.routers())
    edges_to_add = [e for e in g_l3.edges()
                    if e.src in g_bgp and e.dst in g_bgp]
    g_bgp.add_edges_from(edges_to_add, bidirectional=True)
    ank_utils.copy_int_attr_from(g_l3, g_bgp, "multipoint")

    # remove ibgp links

    """TODO: remove up to here once compiler updated"""
    ank_utils.copy_attr_from(
        g_in, g_bgp, "custom_config_bgp", dst_attr="custom_config")

    # log.info("Building eBGP")
    ebgp_nodes = [d for d in g_bgp if any(
        edge.type == 'ebgp' for edge in d.edges())]
    g_bgp.update(ebgp_nodes, ebgp=True)

    for ebgp_edge in g_bgp.edges(type="ebgp"):
        for interface in ebgp_edge.interfaces():
            interface.ebgp = True

    for edge in g_bgp.edges(type='ibgp'):
        # TODO: need interface querying/selection. rather than hard-coded ids
        # TODO: create a new port (once API allows) rarher than binding to
        # loopback zero
        edge.bind_interface(edge.src, 0)

    # TODO: need to initialise interface zero to be a loopback rather than physical type
    # TODO: wat is this for?
    for node in g_bgp:
        for interface in node.interfaces():
            interface.multipoint = any(e.multipoint for e in interface.edges())

    # log.info("Building iBGP")
    build_ibgp(anm)
    build_ibgp_v4(anm)
    build_ibgp_v6(anm)



class NmGraphData(object):

    """API to access link in network"""

    def __init__(self, anm, overlay_id):

        # Set using this method to bypass __setattr__

        object.__setattr__(self, 'anm', anm)
        object.__setattr__(self, 'overlay_id', overlay_id)

    def __repr__(self):
        """"""

        return 'Data for (%s, %s)' % (self.anm, self.overlay_id)

    def dump(self):
        """"""

        print str(self._graph.graph)

    @property
    def _graph(self):

        # access underlying graph for this NmNode

        return self.anm.overlay_nx_graphs[self.overlay_id]

    def __getattr__(self, key):
        return self._graph.graph.get(key)

    def __setattr__(self, key, val):
        self._graph.graph[key] = val

    def __getitem__(self, key):
        """"""

        return self._graph.graph.get(key)

    def __setitem__(self, key, val):
        """"""

        self._graph.graph[key] = val



from autonetkit.topologies.mixed_topology import mixed as mixed
from autonetkit.topologies.multi_edge_topology import multi_edge as multi_edge
from autonetkit.topologies.multi_as_topology import multi_as as multi_as
from autonetkit.topologies.four_chain_topology import four_chain as four_chain
from autonetkit.topologies.house_topology import house as house


from autonetkit.anm.network_model import NetworkModel as NetworkModel
from autonetkit.anm.graph import NmGraph as NmGraph
from autonetkit.anm.node import NmNode as NmNode
from autonetkit.anm.edge import NmEdge as NmEdge
from autonetkit.anm.interface import NmPort as NmPort